#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2022, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################################################################
from cnspy_spatial_csv_formats.EstimationErrorType import EstimationErrorType


# TODO: Unftly, TrajectoryErrorType cannot inherit from EstimationErrorType due to restrictions in Enum!
class TrajectoryErrorType:
    err_type = None # EstimationErrorType

    def __init__(self, err_type=EstimationErrorType.type1):
        self.err_type = err_type

    def __str__(self):
        return self.err_type.__str__()

    def is_local_pose(self):
        return self.err_type == EstimationErrorType.type1

    def is_global_pose(self):
        return self.err_type == EstimationErrorType.type2

    def is_local_pose_inv(self):
        return self.err_type == EstimationErrorType.type3

    def is_global_pose_inv(self):
        return self.err_type == EstimationErrorType.type4

    def is_global_p_local_q(self):
        return self.err_type == EstimationErrorType.type5

    def is_global_p_global_q(self):
        return self.err_type == EstimationErrorType.type6

    def is_local_rotation_error(self):
        if self.err_type == EstimationErrorType.type1 or \
           self.err_type == EstimationErrorType.type3 or \
           self.err_type == EstimationErrorType.type5:
            return True

        return False

    def is_local_position_error(self):
        if self.err_type == EstimationErrorType.type1 or \
                self.err_type == EstimationErrorType.type3 or \
                self.err_type == EstimationErrorType.type5:
            return True
        return False

    def error_def(self):
        p_err_text = ''
        R_err_text = ''
        if self.err_type == EstimationErrorType.type5:
            p_err_text = '(p_EST - p_GT)'
            R_err_text = '(inv(R_GT) * R_EST)'
        elif self.err_type == EstimationErrorType.type6:
            p_err_text = '(p_EST - p_GT)'
            R_err_text = '(R_GT * inv(R_EST))'
        elif self.err_type == EstimationErrorType.type1:
            p_err_text = '(R_EST^(T)(p_EST - p_GT))'
            R_err_text = '(inv(R_EST) * R_GT)'
        elif self.err_type == EstimationErrorType.type2:
            p_err_text = '(p_GT - R_ERR*p_EST)'
            R_err_text = '(R_GT * inv(R_EST))'
        return [p_err_text, R_err_text]

    @staticmethod
    def local_pose():
        return TrajectoryErrorType(err_type=EstimationErrorType.type1)

    @staticmethod
    def global_pose():
        return TrajectoryErrorType(err_type=EstimationErrorType.type2)

    @staticmethod
    def global_p_local_q():
        return TrajectoryErrorType(err_type=EstimationErrorType.type5)

    @staticmethod
    def global_p_global_q():
        return TrajectoryErrorType(err_type=EstimationErrorType.type6)

    @staticmethod
    def list():
        return EstimationErrorType.list()


#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2022, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################################################################
import os
from sys import version_info

import numpy as np
import pandas
import matplotlib.pyplot as plt

from cnspy_csv2dataframe.PoseTypedStamped2DataFrame import PoseTypedStamped2DataFrame
from cnspy_csv2dataframe.TUMCSV2DataFrame import TUMCSV2DataFrame
from cnspy_csv2dataframe.CSV2DataFrame import CSV2DataFrame
from cnspy_spatial_csv_formats.CSVSpatialFormatType import CSVSpatialFormatType
from cnspy_spatial_csv_formats.EstimationErrorType import EstimationErrorType
from cnspy_trajectory.SpatialConverter import SpatialConverter
from cnspy_trajectory.PlotLineStyle import PlotLineStyle
from cnspy_trajectory.Trajectory import Trajectory
from cnspy_trajectory.TrajectoryErrorType import TrajectoryErrorType
from cnspy_trajectory.TrajectoryPlotUtils import TrajectoryPlotUtils, TrajectoryPlotConfig


class TrajectoryError(Trajectory):
    # p_vec: position error vector defined via "error_type" [m]
    # q_vec: rotation error defined via "error_type"  [quaternion]
    format_type = CSVSpatialFormatType.PoseTypedStamped
    traj_err_type = None # TrajectoryErrorType
    scale = 1.0          # position scale factor

    # private metrics:
    # force access to ARMSE via method get_ARMSE()!
    __p_RMSE_vec = None       # norm of position error vector [m]
    __q_RMSE_deg_vec = None   # rotation error angle (phi of axang) [deg]
    __ARMSE_p = None          # [m]
    __ARMSE_q_deg = None      # [deg]

    def __init__(self, t_vec=None, p_vec=None, q_vec=None, scale=1.0,
                 traj_err_type=TrajectoryErrorType(err_type=EstimationErrorType.type1),
                 df=None, fn=None):
        Trajectory.__init__(self, t_vec=t_vec, p_vec=p_vec, q_vec=q_vec)
        self.traj_err_type = traj_err_type
        self.scale = scale

        if df is not None:
            self.load_from_DataFrame(df)
        elif fn is not None:
            self.load_from_CSV(fn)
        pass

    def get_ARMSE(self):
        if self.__ARMSE_p is None or self.__ARMSE_q_deg is None:
            self.compute_ARMSE()
        return self.__ARMSE_p, self.__ARMSE_q_deg

    def compute_ARMSE(self):
        if self.__p_RMSE_vec is None or self.__q_RMSE_deg_vec is None:
            self.__p_RMSE_vec = TrajectoryError.compute_RMSE_p_vec(self.p_vec)
            a, self.__q_RMSE_deg_vec = TrajectoryError.compute_RMSE_q_vec(self.q_vec)

        self.__ARMSE_p = np.mean(self.__p_RMSE_vec)
        self.__ARMSE_q_deg = np.mean(self.__q_RMSE_deg_vec)

    @staticmethod
    def compute_RMSE_p_vec(p_err_arr):
        return np.sqrt(np.sum(p_err_arr ** 2, 1))

    @staticmethod
    def compute_RMSE_q_vec(q_err_arr):
        length = np.shape(q_err_arr)[0]
        rmse_rad_vec = np.zeros((length, 1))
        for i in range(length):
            R_SO3 = SpatialConverter.HTMQ_quaternion_to_SO3(q_err_arr[i, :])
            # Eq. 24 in Zhang and Scaramuzza [1]
            # Bi-invariate metric for rotations (length of geodesic on unit-sphere from identity element) [3]
            [theta, v] = R_SO3.angvec()
            rmse_rad_vec[i, :] = abs(theta)

        rmse_deg_vec = np.rad2deg(rmse_rad_vec)
        return rmse_rad_vec, rmse_deg_vec

    # overriding abstract method
    def load_from_DataFrame(self, df, fmt_type=None):
        assert (isinstance(df, pandas.DataFrame))
        if fmt_type is None:
            fmt_type = CSV2DataFrame.identify_format(dataframe=df)
        else:
            assert (isinstance(fmt_type, CSVSpatialFormatType))

        est_err_type = EstimationErrorType.none
        if fmt_type == CSVSpatialFormatType.PoseStamped:
            self.t_vec, self.p_vec, self.q_vec = \
                TUMCSV2DataFrame.from_DataFrame(df)
            self.scale = 1.0
        elif fmt_type == CSVSpatialFormatType.PoseTypedStamped:
            self.t_vec, self.p_vec, self.q_vec, scale_vec, est_err_type_vec = \
                PoseTypedStamped2DataFrame.from_DataFrame(df)
            self.scale = scale_vec[0]
            est_err_type = EstimationErrorType(est_err_type_vec[0])

        self.traj_err_type = TrajectoryErrorType(est_err_type)
        self.format_type = fmt_type

    # overriding abstract method
    def to_DataFrame(self):
        est_err_type_vec = np.repeat(str(self.traj_err_type.err_type), self.num_elems(), axis=0)
        scale_vec = np.repeat(self.scale, self.num_elems(), axis=0)
        return PoseTypedStamped2DataFrame.to_DataFrame(self.t_vec, self.p_vec, self.q_vec, scale_vec, est_err_type_vec)

    # overriding abstract method
    def clone(self):
        return TrajectoryError(t_vec=self.t_vec.copy(), p_vec=self.p_vec.copy(), q_vec=self.q_vec.copy(),
                              scale=self.scale, traj_err_type=self.traj_err_type)

    ########### PLOTTING #################
    def plot_p_err(self, cfg=TrajectoryPlotConfig(), fig=None, ax=None):
        if fig is None:
            fig = plt.figure(figsize=(20, 15), dpi=int(cfg.dpi))
        if ax is None:
            ax = fig.add_subplot(111)


        self.ax_plot_pos(ax=ax, cfg=cfg)

        [p_err_text, R_err_text] = self.traj_err_type.error_def()
        ARMSE_p, ARMSE_q_deg = self.get_ARMSE()
        ax.set_ylabel(p_err_text + ' ARMSE ={:.2f} [m]'.format(ARMSE_p))

        TrajectoryPlotConfig.show_save_figure(cfg, fig)
        return fig, ax

    def plot_rpy_err(self, cfg=TrajectoryPlotConfig(), fig=None, ax=None):

        if ax is None:
            if fig is None:
                fig = plt.figure(figsize=(20, 15), dpi=int(cfg.dpi))
            ax = fig.add_subplot(111)
        self.ax_plot_rpy(ax=ax, cfg=cfg)

        [p_err_text, R_err_text] = self.traj_err_type.error_def()
        ARMSE_p, ARMSE_q_deg = self.get_ARMSE()
        if cfg.radians:
            ax.set_ylabel(R_err_text + ' ARMSE ={:.2f} [rad]'.format(np.deg2rad(ARMSE_q_deg)))
        else:
            ax.set_ylabel(R_err_text + ' ARMSE ={:.2f} [deg]'.format(ARMSE_q_deg))

        TrajectoryPlotConfig.show_save_figure(cfg, fig)
        return fig, ax



    @staticmethod
    def plot_pose_err(traj_est, traj_err, fig=None, cfg=None, angles=False, plot_rpy=False, traj_gt=None):
        assert(isinstance(traj_err, TrajectoryError))
        assert(isinstance(traj_est, Trajectory))
        assert (isinstance(cfg, TrajectoryPlotConfig))

        if fig is None:
            fig = plt.figure(figsize=(20, 15), dpi=int(cfg.dpi))

        # create 2x2 grid
        ax1 = fig.add_subplot(221)
        ax2 = fig.add_subplot(222)
        ax3 = fig.add_subplot(223)
        ax4 = fig.add_subplot(224)

        # Error type text:
        p_err_text, R_err_text = traj_err.traj_err_type.error_def()

        traj_est.ax_plot_pos(ax=ax1, cfg=cfg)
        if traj_gt:
            assert (isinstance(traj_gt, Trajectory))
            traj_gt.ax_plot_pos(ax=ax1, cfg=cfg, y_label_prefix='true ',
                                colors=['darkred', 'darkgreen', 'darkblue'], labels=['gt_x', 'gt_y', 'gt_z'],
                                ls=PlotLineStyle(linestyle='-.', linewidth=0.5))
        traj_err.ax_plot_pos(ax=ax2, cfg=cfg)

        ax1.set_ylabel('position est [m]')
        ax2.set_ylabel(p_err_text + 'position err [m]')

        if angles:
            traj_est.ax_plot_rpy(ax=ax3, cfg=cfg)
            if traj_gt:
                traj_gt.ax_plot_rpy(ax=ax3, cfg=cfg, colors=['darkred', 'darkgreen', 'darkblue'],
                                    ls=PlotLineStyle(linestyle='-.', linewidth=0.5))

            if plot_rpy:
                traj_err.ax_plot_rpy(ax=ax4, cfg=cfg)
            else:
                traj_err.ax_plot_angle(ax=ax4, cfg=cfg)

            if cfg.radians:
                ax3.set_ylabel('rotation est [rad]')
                ax4.set_ylabel(R_err_text + 'rotation err [rad]')
            else:
                ax3.set_ylabel('rotation est [deg]')
                ax4.set_ylabel(R_err_text + 'rotation err [deg]')
        else:
            traj_est.ax_plot_q(ax=ax3, cfg=cfg)
            if traj_gt:
                traj_gt.ax_plot_q(ax=ax3, cfg=cfg, colors=['k', 'k', 'k', 'k'],
                                  ls=PlotLineStyle(linestyle='-.', linewidth=0.5))
            traj_err.ax_plot_q(ax=ax4, cfg=cfg)
            ax4.set_ylabel(R_err_text + ' quaternion')

        ax1.grid(visible=True)
        ax2.grid(visible=True)
        ax3.grid(visible=True)
        ax4.grid(visible=True)

        TrajectoryPlotConfig.show_save_figure(cfg, fig)

        return fig, ax1, ax2, ax3, ax4



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2020, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Requirements:
# numpy, matplotlib
########################################################################################################################
import numpy as np
from matplotlib import pyplot as plt


########################################################################################################################
# by Mateen Ulhaq
# ref: https://stackoverflow.com/a/50664367

def set_axes_equal(ax=plt.Axes):
    """Set 3D plot axes to equal scale.

    Make axes of 3D plot have equal scale so that spheres appear as
    spheres and cubes as cubes.  Required since `ax.axis('equal')`
    and `ax.set_aspect('equal')` don't work on 3D.
    """
    limits = np.array([
        ax.get_xlim3d(),
        ax.get_ylim3d(),
        ax.get_zlim3d(),
    ])
    origin = np.mean(limits, axis=1)
    radius = 0.5 * np.max(np.abs(limits[:, 1] - limits[:, 0]))
    _set_axes_radius(ax, origin, radius)


def _set_axes_radius(ax, origin, radius):
    x, y, z = origin
    ax.set_xlim3d([x - radius, x + radius])
    ax.set_ylim3d([y - radius, y + radius])
    ax.set_zlim3d([z - radius, z + radius])

########################################################################################################################



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2020, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Requirements:
# enum
########################################################################################################################
import math
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # <--- This is important for 3d plotting
# from mpl_toolkits.mplot3d import Axes3D  # <--- This is important for 3d plotting (copy, if accidentally auto-removed)
from spatialmath import base, SE3

import cnspy_numpy_utils.numpy_statistics
from cnspy_csv2dataframe.TUMCSV2DataFrame import TUMCSV2DataFrame
from cnspy_numpy_utils.numpy_statistics import numpy_statistics
from cnspy_trajectory.TrajectoryBase import TrajectoryBase
from cnspy_trajectory.PlotLineStyle import PlotLineStyle
from cnspy_trajectory.TrajectoryPlotConfig import TrajectoryPlotConfig
from cnspy_trajectory.TrajectoryPlotTypes import TrajectoryPlotTypes
from cnspy_trajectory.TrajectoryPlotUtils import TrajectoryPlotUtils

from cnspy_trajectory.SpatialConverter import SpatialConverter
from cnspy_trajectory.pyplot_utils import set_axes_equal


class Trajectory(TrajectoryBase):
    #t_vec = None
    p_vec = None
    q_vec = None
    dist = None  # cache

    def __init__(self, t_vec=None, p_vec=None, q_vec=None, df=None, fn=None):
        """
            CTOR expects either a pandas.DataFrame or a (timestamp + position + quaternion) matrix

            INPUT:
            t_vec -- [Nx1] matrix, containing N timestamps
            p_vec -- [Nx3] matrix, containing x,y,z positions
            q_vec -- [Nx4] matrix, containing quaternions [x,y,z,w]
            df -- pandas.DataFrame holding a table with ['t', 'tx', 'ty', 'tz','qx', 'qy', 'qz', 'qw']
        """
        TrajectoryBase.__init__(self)
        if df is not None:
            self.load_from_DataFrame(df)
        elif fn is not None:
            self.load_from_CSV(fn=fn)
        elif t_vec is not None and p_vec is not None and q_vec is not None:
            if t_vec.ndim == 1:
                t_vec = np.array([t_vec])

            t_rows, t_cols = t_vec.shape
            p_rows, p_cols = p_vec.shape
            q_rows, q_cols = q_vec.shape
            assert (t_rows == p_rows)
            assert (t_rows == q_rows)
            assert (t_cols == 1)
            assert (p_cols == 3)
            assert (q_cols == 4)

            self.t_vec = t_vec
            self.p_vec = p_vec
            self.q_vec = q_vec

    # overriding abstract method
    def subsample(self, step=None, num_max_points=None, verbose=False):
        sparse_indices = TrajectoryBase.subsample(self, step=step, num_max_points=num_max_points, verbose=verbose)
        self.p_vec = self.p_vec[sparse_indices]
        self.q_vec = self.q_vec[sparse_indices]
        return sparse_indices

    # overriding abstract method
    def sample(self, indices_arr, verbose=False):
        TrajectoryBase.sample(self, indices_arr=indices_arr)
        self.p_vec = self.p_vec[indices_arr]
        self.q_vec = self.q_vec[indices_arr]

    # overriding abstract method
    def clone(self):
        return Trajectory(t_vec=self.t_vec.copy(), p_vec=self.p_vec.copy(), q_vec=self.q_vec.copy())

    # overriding abstract method
    def load_from_DataFrame(self, df, fmt_type=None):
        self.t_vec, self.p_vec, self.q_vec = TUMCSV2DataFrame.from_DataFrame(data_frame=df)

    # overriding abstract method
    def to_DataFrame(self):
        return TUMCSV2DataFrame.to_DataFrame(self.t_vec, self.p_vec, self.q_vec)

    def get_distance(self):
        if self.p_vec is not None:
            if self.dist is not None:
                return self.dist
            else:
                self.dist = Trajectory.distance(self.p_vec)
                return self.dist
        else:
            return 0

    def get_rot_arr(self):
        rot_arr = np.zeros(shape=[self.num_elems(), 3, 3])
        for i in range(self.num_elems()):
            rot_arr[i, :, :] = SpatialConverter.HTMQ_quaternion_to_rot(self.q_vec[i,:])

        return rot_arr

    def set_rot_arr(self, rot_arr):
        assert rot_arr.shape[0] == self.num_elems()
        assert rot_arr.shape[1] == 3
        assert rot_arr.shape[2] == 3

        for i in range(self.num_elems()):
            self.q_vec[i,:] = SpatialConverter.SO3_to_HTMQ_quaternion(rot_arr[i, :, :])


    def get_accumulated_distances(self):
        return Trajectory.distances_from_start(self.p_vec)

    def get_rpy_vec(self):
        rpy_vec = np.zeros(np.shape(self.p_vec))
        for i in range(np.shape(self.p_vec)[0]):
            q = SpatialConverter.HTMQ_quaternion_to_Quaternion(self.q_vec[i, :])
            rpy_vec[i, :] = q.unit().rpy(order='zyx')

        return rpy_vec

    def get_angle_axis_vec(self, unit='rad'):
        phi_vec = np.zeros([np.shape(self.p_vec)[0], 1])
        axis_vec = np.zeros([np.shape(self.p_vec)[0], 3])
        for i in range(np.shape(self.p_vec)[0]):
            q = SpatialConverter.HTMQ_quaternion_to_Quaternion(self.q_vec[i, :])
            phi, u_vec = q.angvec(unit=unit)
            phi_vec[i, :] = phi
            axis_vec[i, :] = u_vec

        return phi_vec, axis_vec

    # TODO: rename parameters and method name. Indicate that it is a local to global transformation (left multiplied)
    #  and the members of trajectory should emphasis the relation between Body and World/Global
    def transform(self, scale=1.0, p_GN_in_G=np.zeros((3,)), R_GN=np.identity(3)):
        p_GB_in_G_arr = np.zeros(np.shape(self.p_vec))
        q_GB_arr = np.zeros(np.shape(self.q_vec))

        T_GN = SpatialConverter.p_R_to_SE3(p_GN_in_G, R_GN)
        for i in range(np.shape(self.p_vec)[0]):
            T_NB = SpatialConverter.p_q_HTMQ_to_SE3(scale * self.p_vec[i, :], self.q_vec[i, :])
            # T_AC = T_AB * T_BC
            # p_AC_in_A = p_AB_in_A + R_AB * p_BC_in_B
            # R_AC = R_AB * R_BC
            p_GB_in_G, q_GB = SpatialConverter.SE3_to_p_q_HTMQ(T_GN * T_NB)
            q_GB_arr[i, :] = q_GB
            p_GB_in_G_arr[i, :] = p_GB_in_G

        self.p_vec = p_GB_in_G_arr
        self.q_vec = q_GB_arr


    def print_statistics(self, file=None):
        metrics_pos = numpy_statistics(self.p_vec)
        metrics_rpy = numpy_statistics(self.get_rpy_vec())
        distance = self.get_distance()
        cnspy_numpy_utils.numpy_statistics.print_statistics(metrics_pos, desc='xyz', file=file)
        cnspy_numpy_utils.numpy_statistics.print_statistics(metrics_rpy, desc='rpy', file=file)
        if file is not None:
            print("distance   %fm" % (distance), file=file)
        else:
            print("distance   %fm" % (distance))

        metrics = {'xyz' : metrics_pos, 'rpy' : metrics_rpy, 'dist' : distance}
        return metrics

    @staticmethod
    def distances_from_start(p_vec):
        distances = np.diff(p_vec[:, 0:3], axis=0)
        # Euclidean distances
        distances = np.sqrt(np.sum(np.multiply(distances, distances), axis=1))
        distances = np.cumsum(distances)
        distances = np.concatenate(([0], distances))
        return distances

    @staticmethod
    def distance(p_vec):
        accum_distances = Trajectory.distances_from_start(p_vec)
        return accum_distances[-1]

    ####################################################################################################################
    ######### PLOTTING #################################################################################################
    ####################################################################################################################
    def get_pos_data(self, cfg):
        assert (isinstance(cfg, TrajectoryPlotConfig))

        ts = self.t_vec
        xs = self.p_vec[:, 0]
        ys = self.p_vec[:, 1]
        zs = self.p_vec[:, 2]
        dist_vec = self.get_accumulated_distances()

        if cfg.white_list:
            print("white list args: " + str(cfg.white_list))
        if any([flag == 'x' for flag in cfg.white_list]):
            xs = []
            print("clear xs")
        if any([flag == 'y' for flag in cfg.white_list]):
            ys = []
            print("clear ys")
        if any([flag == 'z' for flag in cfg.white_list]):
            zs = []
            print("clear zs")

        if not (len(xs) and isinstance(xs[0], np.float64) and not math.isnan(xs[0])):
            xs = []
        if not (len(ys) and isinstance(ys[0], np.float64) and not math.isnan(ys[0])):
            ys = []
        if not (len(zs) and isinstance(zs[0], np.float64) and not math.isnan(zs[0])):
            zs = []

        if cfg.scale and cfg.scale != 1.0:
            scale = float(cfg.scale)
            xs *= scale
            ys *= scale
            zs *= scale

        return ts, xs, ys, zs, dist_vec

    def get_q_data(self, cfg):
        assert (isinstance(cfg, TrajectoryPlotConfig))

        ts = self.t_vec
        qx = self.q_vec[:, 0]
        qy = self.q_vec[:, 1]
        qz = self.q_vec[:, 2]
        qw = self.q_vec[:, 3]
        dist_vec = self.get_accumulated_distances()

        if cfg.white_list:
            print("white list args: " + str(cfg.white_list))
        if any([flag == 'qx' for flag in cfg.white_list]):
            qx = []
            print("clear qx")
        if any([flag == 'qy' for flag in cfg.white_list]):
            qy = []
            print("clear qy")
        if any([flag == 'qz' for flag in cfg.white_list]):
            qz = []
            print("clear qz")
        if any([flag == 'qw' for flag in cfg.white_list]):
            qw = []
            print("clear qw")

        if not (len(qx) and isinstance(qx[0], np.float64) and not math.isnan(qx[0])):
            qx = []
        if not (len(qy) and isinstance(qy[0], np.float64) and not math.isnan(qy[0])):
            qy = []
        if not (len(qz) and isinstance(qz[0], np.float64) and not math.isnan(qz[0])):
            qz = []
        if not (len(qw) and isinstance(qw[0], np.float64) and not math.isnan(qz[0])):
            qw = []
        return ts, qx, qy, qz, qw, dist_vec

    def get_rpy_data(self, cfg):
        assert (isinstance(cfg, TrajectoryPlotConfig))

        rpy_vec = self.get_rpy_vec()

        rpy_vec = np.unwrap(rpy_vec, axis=0)
        if not cfg.radians:
            rpy_vec = np.rad2deg(rpy_vec)

        ts = self.t_vec
        rs = rpy_vec[:, 0]
        ps = rpy_vec[:, 1]
        ys = rpy_vec[:, 2]
        dist_vec = self.get_accumulated_distances()

        if cfg.white_list:
            print("white list args: " + str(cfg.white_list))
        if any([flag == 'roll' for flag in cfg.white_list]):
            rs = []
            print("clear rs")
        if any([flag == 'pitch' for flag in cfg.white_list]):
            ps = []
            print("clear ps")
        if any([flag == 'yaw' for flag in cfg.white_list]):
            ys = []
            print("clear ys")

        if not (len(rs) and isinstance(rs[0], np.float64) and not math.isnan(rs[0])):
            rs = []
        if not (len(ps) and isinstance(ps[0], np.float64) and not math.isnan(ps[0])):
            ps = []
        if not (len(ys) and isinstance(ys[0], np.float64) and not math.isnan(ys[0])):
            ys = []

        return ts, rs, ps, ys, dist_vec

    ###### PLOT-AXIS
    def ax_plot_pos(self, ax, cfg,
                    x_label_prefix='',
                    y_label_prefix='',
                    colors=['r', 'g', 'b'],
                    labels=['x', 'y', 'z'],
                    ls=PlotLineStyle()):

        assert (isinstance(cfg, TrajectoryPlotConfig))

        ts, xs, ys, zs, dist_vec = self.get_pos_data(cfg)

        TrajectoryPlotUtils.ax_plot_pos(ax=ax, ts=ts, xs=xs, ys=ys, zs=zs, dist_vec=dist_vec,
                                        x_label_prefix=x_label_prefix, y_label_prefix=y_label_prefix,
                                        plot_type=cfg.plot_type,
                                        relative_time=cfg.relative_time,
                                        colors=colors, labels=labels, ls=ls)

    def ax_plot_pos_distance(self, ax, cfg,
                    x_label_prefix='',
                    y_label_prefix='',
                    colors=['r'],
                    labels=['norm2([x,y,z])'],
                    ls=PlotLineStyle()):

        assert (isinstance(cfg, TrajectoryPlotConfig))

        ts, xs, ys, zs, dist_vec = self.get_pos_data(cfg)
        ds = np.linalg.norm([xs, ys, zs], axis=0)
        x_arr = TrajectoryPlotUtils.ax_x_linespace(ax=ax, ts=ts, dist_vec=dist_vec,
                                                   relative_time=cfg.relative_time, plot_type=cfg.plot_type,
                                                   x_label_prefix=x_label_prefix)
        ax.set_ylabel(y_label_prefix + 'eucl. distance [m]')
        TrajectoryPlotUtils.ax_plot_n_dim(ax, x_arr, ds, colors=[colors[0]], labels=[labels[0]], ls=ls)



    def ax_plot_rpy(self, ax, cfg,
                    x_label_prefix='',
                    y_label_prefix='Rz(y)Ry(p)Rx(r)',
                    colors=['r', 'g', 'b'],
                    labels=['x', 'y', 'z'],
                    ls=PlotLineStyle()):
        assert (isinstance(cfg, TrajectoryPlotConfig))

        ts, xs, ys, zs, dist_vec = self.get_rpy_data(cfg)

        TrajectoryPlotUtils.ax_plot_rpy(ax=ax,ts=ts, xs=xs, ys=ys, zs=zs, dist_vec=dist_vec,
                                        x_label_prefix=x_label_prefix, y_label_prefix=y_label_prefix,
                                        plot_type=cfg.plot_type,
                                        radians=cfg.radians,
                                        relative_time=cfg.relative_time,
                                        colors=colors, labels=labels, ls=ls)

    def ax_plot_q(self, ax, cfg,
                    x_label_prefix='',
                    y_label_prefix='',
                    colors=['r', 'g', 'b', 'k'],
                    labels=['qx', 'qy', 'qz', 'qw'],
                    ls=PlotLineStyle()):
        assert (isinstance(cfg, TrajectoryPlotConfig))

        ts, qx, qy, qz, qw, dist_vec = self.get_q_data(cfg)

        TrajectoryPlotUtils.ax_plot_q(ax=ax, ts=ts, qx=qx, qy=qy, qz=qz, qw=qw, dist_vec=dist_vec,
                                      x_label_prefix=x_label_prefix,
                                      y_label_prefix=y_label_prefix,
                                      plot_type=cfg.plot_type,
                                      relative_time=cfg.relative_time,
                                      colors=colors, labels=labels, ls=ls)

    def ax_plot_angle(self, ax, cfg,
                      x_label_prefix='',
                      y_label_prefix='angle',
                      colors=['r'],
                      labels=['phi'], ls=PlotLineStyle()):
        assert (isinstance(cfg, TrajectoryPlotConfig))

        unit = 'deg'
        if cfg.radians:
            unit='rad'
        phi_vec, axis_vec = self.get_angle_axis_vec(unit=unit)
        dist_vec = self.get_distance()
        ts  = self.t_vec

        TrajectoryPlotUtils.ax_plot_angle(ax=ax, phi_vec=phi_vec, ts=ts, dist_vec=dist_vec, radians=cfg.radians,
                                          x_label_prefix=x_label_prefix,
                                          y_label_prefix=y_label_prefix,
                                          plot_type=cfg.plot_type,
                                          relative_time=cfg.relative_time,
                                          colors=colors, labels=labels, ls=ls)

    def ax_plot_frames_3D(self, ax, cfg, plot_origin=True, origin_name="0", num_markers=0):
        assert (isinstance(cfg, TrajectoryPlotConfig))

        if plot_origin and not self.is_empty():
            T = SE3(self.p_vec[0, 0], self.p_vec[0, 1], self.p_vec[0, 2])
            base.trplot(T.A, axes=ax, frame=origin_name, rviz=True, length=1, width=0.2, block=False)

        if num_markers > 0 and not self.is_empty():
            for i in range(1, self.num_elems(), max(1, int(self.num_elems() / num_markers))):
                T_i = SpatialConverter.p_q_HTMQ_to_SE3(self.p_vec[i, :], self.q_vec[i, :])
                base.trplot(T_i.A, axes=ax, rviz=True,
                            length=1, width=0.1, block=False)
        pass

    def ax_plot_pos_3D(self, ax, cfg, label='cnspy_trajectory'):
        assert (isinstance(cfg, TrajectoryPlotConfig))

        ts, xs, ys, zs, dist_vec = self.get_pos_data(cfg)

        if cfg.plot_type == TrajectoryPlotTypes.scatter_3D:
            ax.scatter(xs, ys, zs, zdir='z', label=str(label))
        elif cfg.plot_type == TrajectoryPlotTypes.plot_3D:
            ax.plot3D(xs, ys, zs, label=str(label))
        pass

    def plot_pose(self, cfg=TrajectoryPlotConfig(), fig=None, quaternions=False, plot_angle_distance=False,
                  plot_eucl_distance=False):
        assert (isinstance(cfg, TrajectoryPlotConfig))

        if fig is None:
            fig = plt.figure(figsize=(20, 15), dpi=int(cfg.dpi))

        ax1 = fig.add_subplot(211)
        if plot_eucl_distance:
            self.ax_plot_pos_distance(ax=ax1, cfg=cfg)
        else:
            self.ax_plot_pos(ax=ax1, cfg=cfg)
        ax2 = fig.add_subplot(212)

        if plot_angle_distance:
            self.ax_plot_angle(ax=ax2, cfg=cfg)
        else:
            if not quaternions:
                self.ax_plot_rpy(ax=ax2, cfg=cfg)
            else:
                self.ax_plot_q(ax=ax2, cfg=cfg)

        ax1.legend(shadow=True, fontsize='x-small')
        ax1.grid(b=True)
        ax2.legend(shadow=True, fontsize='x-small')
        ax2.grid(b=True)

        TrajectoryPlotConfig.show_save_figure(cfg, fig)
        return fig, ax1, ax2

    def plot_3D(self, fig=None, ax=None, cfg=TrajectoryPlotConfig(), label='traj', num_markers=10):
        assert (isinstance(cfg, TrajectoryPlotConfig))

        if fig is None:
            fig = plt.figure(figsize=(20, 15), dpi=int(cfg.dpi))

        if ax is None:
            ax = fig.add_subplot(111, projection='3d')

        if cfg.title:
            ax.set_title(cfg.title)
        else:
            if cfg.plot_type == TrajectoryPlotTypes.scatter_3D:
                ax.set_title("Scatter Plot")
            else:
                ax.set_title("Plot3D")

        self.ax_plot_pos_3D(ax=ax, cfg=cfg, label=label)
        self.ax_plot_frames_3D(ax=ax, cfg=cfg, plot_origin=True, num_markers=num_markers)

        set_axes_equal(ax)
        ax.legend(shadow=True, fontsize='x-small')
        ax.grid(b=True)
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')
        set_axes_equal(ax)
        TrajectoryPlotConfig.set_view_angle(cfg=cfg, ax=ax)
        TrajectoryPlotConfig.show_save_figure(cfg, fig=fig)

        return fig, ax

    ####################################################################################################################



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2020, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Requirements:
# enum
########################################################################################################################
import os

from cnspy_spatial_csv_formats.ErrorRepresentationType import ErrorRepresentationType
from cnspy_spatial_csv_formats.EstimationErrorType import EstimationErrorType
from cnspy_trajectory.Trajectory import Trajectory
from cnspy_trajectory.TrajectoryError import TrajectoryError
from cnspy_trajectory.TrajectoryErrorType import TrajectoryErrorType
from cnspy_trajectory.TrajectoryEstimated import TrajectoryEstimated
import unittest
import time

from cnspy_trajectory.TrajectoryPlotConfig import TrajectoryPlotConfig
from cnspy_trajectory.TrajectoryPlotTypes import TrajectoryPlotTypes

SAMPLE_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'sample_data')

class TrajectoryError_Test(unittest.TestCase):
    start_time = None

    def start(self):
        self.start_time = time.time()

    def stop(self):
        print("Process time: " + str((time.time() - self.start_time)))

    def load_(self):
        print('loading...')
        fn = str(SAMPLE_DATA_DIR + '/ID1-pose-err.csv')
        obj = TrajectoryError(fn=fn)
        obj.traj_err_type = TrajectoryErrorType(err_type=EstimationErrorType.type1)
        return obj


    def test_plot(self):
        traj_err = self.load_()
        traj_err.save_to_CSV(str(SAMPLE_DATA_DIR + '/results/ID1-pose-err.COPY.csv'))
        cfg = TrajectoryPlotConfig(show=True, close_figure=False,
                                   save_fn=str(SAMPLE_DATA_DIR + '/../../doc/traj_err_pos.png'))
        traj_err.plot_p_err(cfg=cfg)
        cfg = TrajectoryPlotConfig(show=True, close_figure=False,
                                   save_fn=str(SAMPLE_DATA_DIR + '/../../doc/traj_err_rpy.png'))

        traj_err.plot_rpy_err(cfg=cfg)




#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2020, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Requirements:
# enum
########################################################################################################################
from enum import Enum


class TrajectoryPlotTypes(Enum):
    scatter_3D = 'scatter_3D'
    plot_3D = 'plot_3D'
    plot_2D_over_t = 'plot_2D_over_t'
    plot_2D_over_dist = 'plot_2D_over_dist'

    def __str__(self):
        return self.value



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2020, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Requirements:
# numpy, matplotlib, cnspy_csv2dataframe, cnspy_trajectory
########################################################################################################################
import os
import numpy as np
import math
import copy
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # <--- This is important for 3d plotting
# from mpl_toolkits.mplot3d import Axes3D  # <--- This is important for 3d plotting (copy, if accidentally auto-removed)
from spatialmath import base, SE3

from cnspy_csv2dataframe.TUMCSV2DataFrame import TUMCSV2DataFrame
from cnspy_csv2dataframe.CSV2DataFrame import CSV2DataFrame
from cnspy_trajectory.Trajectory import Trajectory
from cnspy_trajectory.TrajectoryEstimated import TrajectoryEstimated
from cnspy_trajectory.TrajectoryError import TrajectoryError
from cnspy_trajectory.TrajectoryPlotConfig import TrajectoryPlotConfig
from cnspy_trajectory.PlotLineStyle import PlotLineStyle
from cnspy_trajectory.TrajectoryPlotTypes import TrajectoryPlotTypes
from cnspy_trajectory.pyplot_utils import set_axes_equal
from cnspy_trajectory.SpatialConverter import SpatialConverter

# TODO: extract some features into dedicated plotting class:
# TODOs PosePlotter:
#  - TODO: ax_plot_n_dim, ax_plot_pos/rpq/q, ax_plot_pos_3D,
#  - TODO: plot_pose
#  - TODO: show_save_figure
#  - TODO: plot 3 sigma bounds
#  - TODO: legend show ARMSE


class TrajectoryPlotter:

    @staticmethod
    def plot_pose_err_cov(traj_est, traj_err, cfg, fig=None, traj_gt=None):
        assert (isinstance(traj_est, TrajectoryEstimated))
        assert (isinstance(traj_err, TrajectoryError))
        assert (isinstance(cfg, TrajectoryPlotConfig))

        # disable functionality
        cfg_ = copy.deepcopy(cfg)
        cfg_.show = False
        cfg_.close = False
        cfg_.save_fn = None
        fig, ax1, ax2, ax3, ax4 = TrajectoryError.plot_pose_err(traj_est=traj_est, traj_err=traj_err,
                                                                cfg=cfg_, fig=fig,
                                                                angles=True, plot_rpy=True,
                                                                traj_gt=traj_gt)
        traj_est.ax_plot_p_sigma(ax2, cfg=cfg_, colors=['darkred', 'darkgreen', 'darkblue'], ls=PlotLineStyle(linestyle='-.', linewidth=0.5))
        traj_est.ax_plot_rpy_sigma(ax4, cfg=cfg_, colors=['darkred', 'darkgreen', 'darkblue'], ls=PlotLineStyle(linestyle='-.', linewidth=0.5))

        TrajectoryPlotConfig.show_save_figure(cfg, fig)
        return fig, ax1, ax2, ax3, ax4

    @staticmethod
    def multi_plot_3D(traj_list, cfg, name_list=None, color_map="gist_rainbow"):
        assert (isinstance(cfg, TrajectoryPlotConfig))
        num_plots = len(traj_list)

        fig = plt.figure(figsize=(20, 15), dpi=int(cfg.dpi))
        ax = fig.add_subplot(111, projection='3d')

        # https://stackoverflow.com/questions/51452112/how-to-fix-cm-spectral-module-matplotlib-cm-has-no-attribute-spectral
        # choose among: https://matplotlib.org/stable/users/explain/colors/colormaps.html#miscellaneous
        cmap = plt.cm.get_cmap(color_map)
        colors = cmap(np.linspace(0.01, 0.99, num_plots))
        ax.set_prop_cycle('color', colors)

        idx = 0
        for traj in traj_list:
            label = "traj" + str(idx)
            if name_list:
                label = name_list[idx]
            traj.ax_plot_pos_3D(ax=ax, cfg=cfg, label=label)
            idx += 1

        ax.legend(shadow=True, fontsize='x-small')
        ax.grid(b=True)
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')

        set_axes_equal(ax)
        TrajectoryPlotConfig.set_view_angle(cfg=cfg, ax=ax)
        plt.draw()
        plt.pause(0.001)

        TrajectoryPlotConfig.show_save_figure(cfg, fig)

        return fig, ax





#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2020, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# adapted from:
# * https://github.com/uzh-rpg/rpg_trajectory_evaluation/blob/master/src/rpg_trajectory_evaluation/align_trajectory.py
#
# Requirements:
# numpy, cnspy_numpy_utils
########################################################################################################################
import numpy as np
from spatialmath import UnitQuaternion, SO3, SE3, Quaternion, base

from cnspy_spatial_csv_formats.ErrorRepresentationType import ErrorRepresentationType
from cnspy_trajectory.SpatialConverter import SpatialConverter
import unittest


class SpatialConverter_Test(unittest.TestCase):
    def test_SO3_to_HTMQ_quaternion(self):
        R1 = SO3.RPY([0, 0, 45], unit='deg')

        print(R1)
        q1 = SpatialConverter.SO3_to_HTMQ_quaternion(R1)
        print(q1)
        q2 = SpatialConverter.SO3_to_HTMQ_quaternion(R1.R)
        print(q2)
        self.assertTrue(((q1 == q2).all()))

        q3 = SpatialConverter.SO3_to_HTMQ_quaternion(np.eye(3))
        print(q3)
        self.assertTrue(((q3 == np.array([0, 0, 0, 1])).all()))

        pass

    def test_HTMQ_quaternion_to_SO3(self):
        q1_TUM = np.array([0, 0,  0.38268343, 0.92387953])
        print(q1_TUM)
        R1 = SpatialConverter.HTMQ_quaternion_to_SO3(q1_TUM)
        print(R1)
        pass

    def test_p_q_HTMQ_to_SE3(self):
        p = np.array([1,2,3])
        q = np.array([0, 0,  0.38268343, 0.92387953])

        SE3_ = SpatialConverter.p_q_HTMQ_to_SE3(p, q)
        print(SE3_)
        self.assertTrue((SE3_.t == np.array([1,2,3]) ).all())
        R1 = SO3.RPY([0, 0, 45], unit='deg')

        diff = SE3_.R - R1.R
        print(diff)
        self.assertTrue( np.allclose(SE3_.R, R1.R))

    def test_SE3_to_HTMQ_p_q(self):
        SE3_ = SE3(1,2,3)
        # SE3 is missing a CTOR for rotation matrices
        SE3_R = SE3.RPY([0, 0, 45], unit='deg')
        SE3_ = SE3_ * SE3_R

        p, q = SpatialConverter.SE3_to_p_q_HTMQ(SE3_)

        print(p)
        print(q)
        self.assertTrue((p == np.array([1, 2, 3])).all())
        self.assertTrue(np.allclose(q, np.array([0, 0,  0.38268343, 0.92387953])))


    def test_rpy2rot(self):
        R_ref = SO3.RPY([0, 0, 45], unit='deg')
        rpy = np.array([0, 0, 45])
        R_ = SpatialConverter.rpy2rot(rpy, unit='deg')
        rpy_ = SpatialConverter.rot2rpy(R_, unit='deg')
        print(R_ref - R_)
        print(rpy - rpy_)
        self.assertTrue(np.allclose(R_ref, R_))
        self.assertTrue(np.allclose(rpy, rpy_))

    def test_theta_q(self):
        q = np.array([0, 0, 0.38268343, 0.92387953])

        theta_q = SpatialConverter.quat2theta_q(q)
        print(theta_q)
        q_ = SpatialConverter.theta_q2quat(theta_q)
        R_ = SpatialConverter.theta_q2rot(theta_q)
        self.assertTrue(np.allclose(q, q_))

    def test_theta_so3(self):
        q = np.array([0, 0, 0.38268343, 0.92387953])

        theta_so3 = SpatialConverter.quat2theta_so3(q)
        print(theta_so3)
        q_ = SpatialConverter.theta_so3_2quat(theta_so3)
        self.assertTrue(np.allclose(q, q_))

        R_ref = SO3.RPY([0, 0, 45], unit='deg').R
        theta_so3 = SpatialConverter.rot2theta_so3(R_ref)
        print(theta_so3)
        R_ = SpatialConverter.theta_so3_2rot(theta_so3)
        self.assertTrue(np.allclose(R_ref, R_))

    def test_theta_R(self):
        R_ref = SO3.RPY([4, 2, 4], unit='deg').R
        theta_R = SpatialConverter.rot2theta_R(R_ref)
        print(theta_R)
        R_ = SpatialConverter.theta_R2rot(theta_R)
        print(R_ref - R_)


    def test_convert_q_vec_to_theta_vec(self):
        len = 10
        q_vec = np.zeros((len, 4))

        for i in range(len):
            q_vec[i] = SpatialConverter.SO3_to_HTMQ_quaternion(SO3.RPY([i, -i, i*2], unit='deg'))

        theta_R_vec = SpatialConverter.convert_q_vec_to_theta_vec(q_vec,
                                                                  rot_err_rep=ErrorRepresentationType.theta_R)
        theta_q_vec = SpatialConverter.convert_q_vec_to_theta_vec(q_vec,
                                                                  rot_err_rep=ErrorRepresentationType.theta_q)
        theta_so3_vec = SpatialConverter.convert_q_vec_to_theta_vec(q_vec,
                                                                    rot_err_rep=ErrorRepresentationType.theta_so3)
        print('\ntheta_R:\n' + str(theta_R_vec))
        print('\ntheta_q:\n' + str(theta_q_vec))
        print('\ntheta_so3:\n' + str(theta_so3_vec))

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2020, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################################################################
import os
import unittest
import math
import numpy as np
from spatialmath import UnitQuaternion, SE3, SO3, Quaternion
import cnspy_trajectory as tr
from cnspy_trajectory.TrajectoryPlotTypes import TrajectoryPlotTypes
from cnspy_trajectory.Trajectory import Trajectory
from cnspy_trajectory.TrajectoryPlotConfig import TrajectoryPlotConfig
from cnspy_trajectory.TrajectoryPlotter import *

SAMPLE_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'sample_data')

class Trajectory_Test(unittest.TestCase):
    def load_trajectory_from_CSV(self):
        print('loading...')
        traj = Trajectory()
        traj.load_from_CSV(fn=str(SAMPLE_DATA_DIR + '/ID1-pose-est.csv'))
        return traj

    def test_load_trajectory_from_CSV(self):
        traj = self.load_trajectory_from_CSV()
        self.assertTrue(traj.num_elems() > 0)

    def test_save_to_CSV(self):
        traj = Trajectory()

        saved = traj.save_to_CSV(str(SAMPLE_DATA_DIR + '/results/empty1.csv'))
        self.assertFalse(saved)

        traj.p_vec = np.array([[0, 0, 0],
                               [1, 0, 0],
                               [2, 0, 0],
                               [3, 0, 0]])
        traj.q_vec = np.array([[0, 0, 0, 1],
                               [0, 0, 0, 1],
                               [0, 0, 0, 1],
                               [0, 0, 0, 1]])
        traj.t_vec = np.array([[0],
                               [1],
                               [2],
                               [3]])
        saved = traj.save_to_CSV(str(SAMPLE_DATA_DIR + '/results/empty2.csv'))
        self.assertTrue(saved)

    def test_get_distance_from_start(self):
        p_vec = np.array([[0, 0, 0],
                          [1, 0, 0],
                          [2, 0, 0],
                          [3, 0, 0]])

        d = Trajectory.distance(p_vec)
        self.assertTrue(math.floor(d - 3.0) == 0)

        p_vec = np.array([[0, 0, 0],
                          [1, 1, 0],
                          [2, 2, 0],
                          [3, 3, 0]])
        d = Trajectory.distance(p_vec)
        d_ = math.sqrt(9 + 9)
        self.assertTrue(math.floor(d - d_) == 0)

        p_vec = np.array([[0, 0, 0],
                          [1, 1, 1],
                          [2, 2, 2],
                          [3, 3, 3]])

        d = Trajectory.distance(p_vec)
        d_ = math.sqrt(9 + 9 + 9)
        self.assertTrue(math.floor(d - d_) == 0)

    def test_get_rpy_vec(self):
        traj = self.load_trajectory_from_CSV()
        rpy_vec = traj.get_rpy_vec()
        self.assertTrue(rpy_vec.shape[0] == traj.num_elems())
        self.assertTrue(rpy_vec.shape[1] == 3)

    def test_get_angle_axis_vec(self):
        traj = self.load_trajectory_from_CSV()
        phi_vec, axis_vec_ = traj.get_angle_axis_vec(unit='rad')
        phi_vec_, axis_vec = traj.get_angle_axis_vec(unit='deg')
        self.assertTrue(axis_vec.shape[0] == traj.num_elems())
        self.assertTrue(axis_vec.shape[1] == 3)
        self.assertTrue(phi_vec.shape[0] == traj.num_elems())
        self.assertTrue(phi_vec.shape[1] == 1)

    def test_transform(self):
        traj = self.load_trajectory_from_CSV()
        self.assertTrue(traj.num_elems() > 0)
        d = traj.get_distance()
        print('distance: ' + str(d))

        R1 = SO3.RPY([0, 0, 45], unit='deg')
        p = np.array([1, 2, 3])
        traj.transform(1, p, R1.R)

        d2 = traj.get_distance()
        print('distance2:  ' + str(d2))

    def test_transform2(self):
        traj = Trajectory()

        saved = traj.save_to_CSV(str(SAMPLE_DATA_DIR + '/results/empty1.csv'))
        self.assertFalse(saved)

        traj.q_vec = np.array([[0, 0, 0, 1],
                               [0, 0, 0, 1],
                               [0, 0, 0, 1],
                               [0, 0, 0, 1],
                               [0, 0, 0, 1],
                               [0, 0, 0, 1],
                               [0, 0, 0, 1]])
        traj.t_vec = np.array([[0],
                               [1],
                               [2],
                               [3], [4], [5], [6]])

        p_GB_in_G_arr = np.array(
            [[0.029, 0.0165, -2.51], [0.0303, 0.3367, -2.511], [0.038, 1.221, -2.516], [0.03115, 1.591, -2.61],
             [0.023, 2.416, -2.819], [-0.03825, 2.79245, -2.82835], [-0.1567, 3.54785, -2.8475]])

        traj.p_vec = p_GB_in_G_arr

        traj2 = Trajectory()
        p_NB_in_N_arr = np.array(
            [[-2.694, -0.049, 0.0205], [-2.696, -0.0475, 0.369], [-2.6935, -0.0435, 1.2225], [-2.623, 0.0175, 1.589],
             [-2.4685, 0.154, 2.403], [-2.4205, 0.123, 2.7725], [-2.317, 0.051, 3.522]])
        traj2.q_vec = np.array([[0, 0, 0, 1],
                               [0, 0, 0, 1],
                               [0, 0, 0, 1],
                               [0, 0, 0, 1],
                               [0, 0, 0, 1],
                               [0, 0, 0, 1],
                               [0, 0, 0, 1]])
        traj2.t_vec = np.array([[0],
                               [1],
                               [2],
                               [3], [4], [5], [6]])
        traj2.p_vec = p_NB_in_N_arr


        cfg=TrajectoryPlotConfig(show=True, close_figure=False, save_fn='')
        #fig, ax = traj.plot_3D(cfg=cfg)
        #traj2.plot_3D(cfg=cfg)

        R_GN = np.array([[ 0.92866884,  0.31819331, -0.19059698],       [ 0.218043  , -0.05263923,  0.97451853],       [ 0.3000524 , -0.94656333, -0.11826423]])
        t_GN = np.array([ 2.69581276,  0.61577898, -1.66060676])

        #R_GN = np.array([[ 0.94404031,  0.12013279, -0.30717424], [-0.30010691, -0.07350927, -0.95106899], [-0.13683473,  0.99003257, -0.03334304]])
        #t_GN = np.array([-3.56530399, -2.3538825 , -0.06128361])
        traj3 = traj2.clone()
        traj3.transform(scale=1, R_GN=R_GN, p_GN_in_G=t_GN)
        TrajectoryPlotter.multi_plot_3D([traj, traj2, traj3], cfg=cfg, name_list=['gt', 'est', 'aligned'])





    def test_plot_3D(self):
        traj = self.load_trajectory_from_CSV()
        cfg=TrajectoryPlotConfig(show=True, close_figure=False, save_fn=str(SAMPLE_DATA_DIR + '/../../doc/plot_3D.png'))
        traj.plot_3D(cfg=cfg)

    def test_plot_pose(self):
        traj = self.load_trajectory_from_CSV()
        cfg=TrajectoryPlotConfig(show=False, close_figure=False, save_fn=str(SAMPLE_DATA_DIR + '/../../doc/pose.png'))
        traj.plot_pose(cfg=cfg)
        traj.plot_pose(cfg=cfg, quaternions=False)
        cfg.radians = False
        traj.plot_pose(cfg=cfg, quaternions=False)

        traj.plot_pose(cfg=cfg, plot_angle_distance=True)
        traj.plot_pose(cfg=cfg, plot_angle_distance=True)

        traj.plot_pose(quaternions=True, cfg=TrajectoryPlotConfig(show=False,
                                                                  close_figure=False,
                                                                  radians=False,
                                                                  plot_type=TrajectoryPlotTypes.plot_2D_over_dist))
        traj.plot_pose(quaternions=True, cfg=TrajectoryPlotConfig(show=True,
                                                                  close_figure=False,
                                                                  radians=False,
                                                                  plot_type=TrajectoryPlotTypes.plot_2D_over_t))
if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2020, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# adapted from:
# * https://github.com/uzh-rpg/rpg_trajectory_evaluation/blob/master/src/rpg_trajectory_evaluation/align_trajectory.py
#
# Requirements:
# numpy, cnspy_numpy_utils
########################################################################################################################
import numpy as np
from spatialmath import UnitQuaternion, SO3, SE3, Quaternion, base

from cnspy_spatial_csv_formats.ErrorRepresentationType import ErrorRepresentationType


class SpatialConverter:
    """
    SpatialConverter for:
    Homogeneous Transformation Matrices and Quaternions (HTMQ) by `Christoph Gohlke <http://www.lfd.uci.edu/~gohlke/>`__,
      Laboratory for Fluorescence Dynamics, University of California, Irvine (https://pypi.org/project/transformations/)

    Problem: the HTMQ quaternion format is non-alphabetic: [x,y,z,w], with w being the scalar part and [x,y,z] the imaginary
    The spatialmath toolbox (https://github.com/petercorke/spatialmath-python) uses [s, v] ([w, x, y, z]),
    s = scalar/real part and v = vector/imaginary part

    In general this convert is required from a historical point, as the spatialmath toolbox is pretty new and would
    require to refactor the whole code base (the [x,y,z,w] format is widely used!)
    """

    def __init__(self):
        pass
    @staticmethod
    def HTMQ_quaternion_identity():
        # [x, y, z, w]
        return np.array([0, 0, 0, 1])

    @staticmethod
    def UnitQuaternion_to_HTMQ_quaternion(q_AB):
        assert (isinstance(q_AB, UnitQuaternion))

        # Convert SO(3) rotation matrix to unit-quaternion
        q_AB_ = base.r2q(q_AB.R)
        indices_inv = [1, 2, 3, 0]
        return q_AB_[indices_inv]

    @staticmethod
    def HTMQ_quaternion_to_Quaternion(q_AB):
        # numpy is inconsistent with shape! as q_AB.shape == (4, 1) fails!
        if not (isinstance(q_AB, np.ndarray) and len(q_AB) == 4):
            raise ValueError('Invalid input')
        indices = [3, 0, 1, 2]
        return UnitQuaternion(q_AB[indices])

    @staticmethod
    def HTMQ_quaternion_to_rot(q_AB):
        # numpy is inconsistent with shape! as q_AB.shape == (4, 1) fails!
        if not (isinstance(q_AB, np.ndarray) and len(q_AB) == 4):
            raise ValueError('Invalid input')
        indices = [3, 0, 1, 2]
        return UnitQuaternion(q_AB[indices]).R

    @staticmethod
    def SO3_to_HTMQ_quaternion(R_AB):
        # returns a np.array representing a unit-quaternion [x,y,z, w], [[v], s];
        # Convert SO(3) rotation matrix to unit-quaternion
        if isinstance(R_AB, SO3):
            return base.r2q(R_AB.R, order="xyzs")
        elif base.isrot(R_AB, check=False):
            return base.r2q(R_AB, order="xyzs")
        else:
            print(str(R_AB))
            raise ValueError('Invalid type')

    @staticmethod
    def HTMQ_quaternion_to_SO3(q_AB):
        # numpy is inconsistent with shape! as q_AB.shape == (4, 1) fails!
        if not (isinstance(q_AB, np.ndarray) and len(q_AB) == 4):
            raise ValueError('Invalid input')
        indices = [3, 0, 1, 2]
        SO3_ = SO3(Quaternion(q_AB[indices]).unit().R)
        return SO3_

    @staticmethod
    def p_q_HTMQ_to_SE3(p_AB_in_A, q_AB):
        if not (isinstance(q_AB, np.ndarray) and len(q_AB) == 4):
            raise ValueError('Invalid input')
        if not (isinstance(p_AB_in_A, np.ndarray) and len(p_AB_in_A) == 3):
            raise ValueError('Invalid input')

        T = np.identity(4)
        R_AB = SpatialConverter.HTMQ_quaternion_to_SO3(q_AB)
        T[0:3, 0:3] = R_AB.R
        T[:3, 3] = p_AB_in_A
        T_AB = SE3(T)
        return T_AB

    @staticmethod
    def p_R_to_SE3(p_AB_in_A, R_AB):
        if not (isinstance(R_AB, np.ndarray) and R_AB.shape == (3, 3)):
            print(str(R_AB))
            raise ValueError('Invalid input')
        if not (isinstance(p_AB_in_A, np.ndarray) and len(p_AB_in_A) == 3):
            print(str(p_AB_in_A))
            raise ValueError('Invalid input')

        T = np.identity(4)
        T[0:3, 0:3] = R_AB
        T[:3, 3] = p_AB_in_A
        T_AB = SE3(T, check=False)
        return T_AB

    @staticmethod
    def SE3_to_p_q_HTMQ(T_AB):
        assert(isinstance(T_AB, SE3))
        R_AB = T_AB.R
        t_AB = T_AB.t

        return t_AB, SpatialConverter.SO3_to_HTMQ_quaternion(R_AB)

    @staticmethod
    def rpy2rot(rpy, unit='rad'):
        return base.rpy2r(roll=rpy[0], pitch=rpy[1], yaw=rpy[2], unit=unit, order='zyx')

    @staticmethod
    def rot2rpy(R, unit='rad'):
        # R = rotz(angles[2]) @ roty(angles[1]) @ rotx(angles[0])
        return base.tr2rpy(T=R, unit=unit, order='zyx')

    @staticmethod
    def quat2theta_q(q_err):
        """
        converts a rotation represented by a quaternion in its small angle approximation
        > S = R * Exp(theta)
        > R(q_err) = R^T * S
        > theta = Log(R(q_err))
        refer to:
        * "Quaternion kinematics for the error-state Kalman filter" by Joan Sola,
            https://arxiv.org/pdf/1711.02508.pdf, chapter 4: Perturbations, derivatives and integrals
        For really small perturbations the following should be a sufficient approximation:
        > R(q) = I + skew(theta)   // Sola EQ. 193
        > theta = unskew(R(q) - I)
        """
        return  2 * (q_err[:3] / q_err[3])

    @staticmethod
    def theta_q2quat(theta_q):
        q_ = Quaternion(s=1, v=theta_q*0.5)
        return SpatialConverter.UnitQuaternion_to_HTMQ_quaternion(q_.unit())

    @staticmethod
    def theta_q2rot(theta_q):
        q_ = Quaternion(s=1, v=theta_q*0.5)
        return q_.unit().R

    @staticmethod
    def quat2theta_so3(q_err):
        # R = expm(skew(theta_so3))
        R_ = SpatialConverter.HTMQ_quaternion_to_SO3(q_err)
        return base.trlog(R_.R, twist=True)

    @staticmethod
    def theta_so3_2quat(theta_so3):
        R_ = base.exp2tr(theta_so3)
        if R_.shape == (4,4):
            R_ = R_[0:3,0:3]
        return SpatialConverter.SO3_to_HTMQ_quaternion(R_)

    @staticmethod
    def rot2theta_so3(R):
        # R = expm(skew(theta_so3))
        return base.trlog(R, twist=True)

    @staticmethod
    def theta_so3_2rot(theta_so3):
        R_ =  base.exp2tr(theta_so3)
        if R_.shape == (4,4):
            R_ = R_[0:3,0:3]
        return R_

    @staticmethod
    def quat2theta_R(q_err):
        # R = I + skew(theta_R)
        R_ = SpatialConverter.HTMQ_quaternion_to_SO3(q_err)
        return base.vex(R_.R - np.eye(3))

    @staticmethod
    def theta_R2rot(theta_R):
        R_ = np.eye(3) + base.skew(theta_R)
        # TODO: normalization of rotation matrix via UnitQuaternion
        q_ = UnitQuaternion(SO3(R_, check=False)).unit()
        return q_.R

    @staticmethod
    def rot2theta_R(R):
        # R = I + skew(theta_R)
        return base.vex(R - np.eye(3))

    @staticmethod
    def convert_q_vec_to_theta_vec(q_vec, rot_err_rep):
        assert (isinstance(rot_err_rep, ErrorRepresentationType))
        # q_vec as vector of  HTMQ_quaternions
        len, q_cols = q_vec.shape
        assert(q_cols == 4)
        theta_vec = np.zeros((len, 3))

        # Converts quaternion to small angle approximations
        if rot_err_rep == ErrorRepresentationType.theta_R:
            for i in range(len):
                theta_vec[i] = SpatialConverter.quat2theta_R(q_vec[i])
        elif rot_err_rep == ErrorRepresentationType.theta_q:
            for i in range(len):
                theta_vec[i] = SpatialConverter.quat2theta_q(q_vec[i])
        elif rot_err_rep == ErrorRepresentationType.theta_so3:
            for i in range(len):
                theta_vec[i] = SpatialConverter.quat2theta_so3(q_vec[i])
        else:
            assert False, 'format is not supported: ' + str(rot_err_rep)

        return theta_vec



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2022, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Requirements:
# numpy, matplotlib, cnspy_csv2dataframe, cnspy_trajectory
########################################################################################################################
import os
import numpy as np
import math
import matplotlib.pyplot as plt
from cnspy_trajectory.TrajectoryPlotConfig import TrajectoryPlotConfig
from cnspy_trajectory.PlotLineStyle import PlotLineStyle
from cnspy_trajectory.TrajectoryPlotTypes import TrajectoryPlotTypes

class TrajectoryPlotUtils:

    @staticmethod
    def ax_plot_n_dim(ax, x_linespace, values,
                      colors=['r', 'g', 'b'],
                      labels=['x', 'y', 'z'], ls=PlotLineStyle()):
        assert len(colors) == len(labels)
        if len(colors) > 1:
            assert len(colors) == values.shape[1]
            for i in range(len(colors)):
                ax.plot(x_linespace, values[:, i],
                        color=colors[i], label=labels[i], linestyle=ls.linestyle, linewidth=ls.linewidth,
                        marker=ls.marker)
        else:
            ax.plot(x_linespace, values, color=colors[0], label=labels[0], linestyle=ls.linestyle,
                    linewidth=ls.linewidth)
    @staticmethod
    def ax_x_linespace(ax, ts=None, dist_vec=None, relative_time=True, plot_type=None, x_label_prefix=''):
        if plot_type == TrajectoryPlotTypes.plot_2D_over_dist:
            x_arr = dist_vec
            ax.set_xlabel(x_label_prefix + 'distance [m]')
        else:
            x_arr = ts
            if relative_time:
                x_arr = x_arr - x_arr[0]
                ax.set_xlabel(x_label_prefix + 'rel. time [sec]')
            else:
                ax.set_xlabel(x_label_prefix + 'time [sec]')
        return x_arr

    @staticmethod
    def ax_plot_pos(ax, ts=None, xs=None, ys=None, zs=None, dist_vec=None,
                    x_label_prefix='',
                    y_label_prefix='',
                    plot_type=TrajectoryPlotTypes.plot_2D_over_dist,
                    relative_time=True,
                    colors=['r', 'g', 'b'], labels=['x', 'y', 'z'],
                    ls=PlotLineStyle()):

        x_arr = TrajectoryPlotUtils.ax_x_linespace(ax=ax, ts=ts, dist_vec=dist_vec,
                                                   relative_time=relative_time, plot_type=plot_type,
                                                   x_label_prefix=x_label_prefix)
        ax.set_ylabel(y_label_prefix + 'position [m]')

        if xs is not None and len(xs):
            TrajectoryPlotUtils.ax_plot_n_dim(ax, x_arr, xs, colors=[colors[0]], labels=[labels[0]], ls=ls)
        if ys is not None and len(ys):
            TrajectoryPlotUtils.ax_plot_n_dim(ax, x_arr, ys, colors=[colors[1]], labels=[labels[1]], ls=ls)
        if zs is not None and len(zs):
            TrajectoryPlotUtils.ax_plot_n_dim(ax, x_arr, zs, colors=[colors[2]], labels=[labels[2]], ls=ls)

    @staticmethod
    def ax_plot_rpy(ax, ts=None, xs=None, ys=None, zs=None, dist_vec=None,
                    x_label_prefix='',
                    y_label_prefix='Rz(y)Ry(p)Rx(r)',
                    plot_type=TrajectoryPlotTypes.plot_2D_over_dist,
                    radians=True,
                    relative_time=True,
                    colors=['r', 'g', 'b'], labels=['x', 'y', 'z'],
                    ls=PlotLineStyle()):

        x_arr = TrajectoryPlotUtils.ax_x_linespace(ax=ax, ts=ts, dist_vec=dist_vec,
                                                   relative_time=relative_time, plot_type=plot_type,
                                                   x_label_prefix=x_label_prefix)
        if radians:
            ax.set_ylabel(y_label_prefix + ' [rad]')
        else:
            ax.set_ylabel(y_label_prefix + ' [deg]')

        if len(xs):
            TrajectoryPlotUtils.ax_plot_n_dim(ax, x_arr, xs, colors=[colors[0]], labels=[labels[0]], ls=ls)
        if len(ys):
            TrajectoryPlotUtils.ax_plot_n_dim(ax, x_arr, ys, colors=[colors[1]], labels=[labels[1]], ls=ls)
        if len(zs):
            TrajectoryPlotUtils.ax_plot_n_dim(ax, x_arr, zs, colors=[colors[2]], labels=[labels[2]], ls=ls)



    @staticmethod
    def ax_plot_q(ax, ts=None, qx=None, qy=None, qz=None, qw=None, dist_vec=None,
                  x_label_prefix='',
                  y_label_prefix='',
                  plot_type=TrajectoryPlotTypes.plot_2D_over_dist,
                  relative_time=True,
                  colors=['r', 'g', 'b', 'k'],
                  labels=['qx', 'qy', 'qz', 'qw'], ls=PlotLineStyle()):

        x_arr = TrajectoryPlotUtils.ax_x_linespace(ax=ax, ts=ts, dist_vec=dist_vec,
                                                   relative_time=relative_time, plot_type=plot_type,
                                                   x_label_prefix=x_label_prefix)
        ax.set_ylabel(y_label_prefix + 'quaternion')

        if len(qx):
            TrajectoryPlotUtils.ax_plot_n_dim(ax, x_arr, qx, colors=[colors[0]], labels=[labels[0]], ls=ls)
        if len(qy):
            TrajectoryPlotUtils.ax_plot_n_dim(ax, x_arr, qy, colors=[colors[1]], labels=[labels[1]], ls=ls)
        if len(qz):
            TrajectoryPlotUtils.ax_plot_n_dim(ax, x_arr, qz, colors=[colors[2]], labels=[labels[2]], ls=ls)
        if len(qw):
            TrajectoryPlotUtils.ax_plot_n_dim(ax, x_arr, qw, colors=[colors[3]], labels=[labels[3]], ls=ls)

    @staticmethod
    def ax_plot_angle(ax, phi_vec, ts=None, dist_vec=None, radians=True,
                      x_label_prefix='',
                      y_label_prefix='angle',
                      plot_type=TrajectoryPlotTypes.plot_2D_over_dist,
                      relative_time=True,
                      colors=['r'],
                      labels=['phi'],
                      ls=PlotLineStyle()):
        x_arr = TrajectoryPlotUtils.ax_x_linespace(ax=ax, ts=ts, dist_vec=dist_vec,
                                                   relative_time=relative_time, plot_type=plot_type,
                                                   x_label_prefix=x_label_prefix)

        if radians:
            ax.set_ylabel(y_label_prefix + '[rad]')
        else:
            ax.set_ylabel(y_label_prefix + '[deg]')

        TrajectoryPlotUtils.ax_plot_n_dim(ax, x_arr, phi_vec, colors=colors,
                                        labels=labels, ls=ls)


#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2020, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################################################################
import math
import os

import numpy as np
import pandas
import matplotlib.pyplot as plt
from spatialmath import base

from cnspy_csv2dataframe.PosOrientWithCovTyped2DataFrame import PosOrientWithCovTyped2DataFrame
from cnspy_csv2dataframe.PoseWithCovTyped2DataFrame import PoseWithCovTyped2DataFrame
from cnspy_trajectory.PlotLineStyle import PlotLineStyle
from cnspy_trajectory.SpatialConverter import SpatialConverter
from cnspy_trajectory.Trajectory import Trajectory
from cnspy_csv2dataframe.PosOrientWithCov2DataFrame import PosOrientWithCov2DataFrame
from cnspy_csv2dataframe.PoseWithCov2DataFrame import PoseWithCov2DataFrame
from cnspy_csv2dataframe.CSV2DataFrame import CSV2DataFrame
from cnspy_csv2dataframe.TUMCSV2DataFrame import TUMCSV2DataFrame
from cnspy_spatial_csv_formats.CSVSpatialFormatType import CSVSpatialFormatType
from cnspy_spatial_csv_formats.CSVSpatialFormat import CSVSpatialFormat
from cnspy_spatial_csv_formats.EstimationErrorType import EstimationErrorType
from cnspy_spatial_csv_formats.ErrorRepresentationType import ErrorRepresentationType
from cnspy_trajectory.TrajectoryPlotConfig import TrajectoryPlotConfig
from cnspy_trajectory.TrajectoryPlotUtils import TrajectoryPlotUtils

# TODO: make TrajectoryEstimatedBase and derive TrajEstPose and TrajEstPosOrient
# TODO: convert locally estimated covariance into global covariance using the
#  adjoint: https://gtsam.org/2021/02/23/uncertainties-part3.html Σ_W = Ad_{T_{WB}} Σ_B Ad_{T_{WB}}^{T}
# TODO: overwrite transform to rotate global covariances!
class TrajectoryEstimated(Trajectory):
    # position uncertainty: 3x3 covariance matrix.
    # The with upper triangular elements are vectorized: 'pxx', 'pxy', 'pxz', 'pyy', 'pyz', 'pzz'
    Sigma_p_vec = None

    # small angle `theta` uncertainty: R = ( eye(3) + skew(theta_R)), a 3x3 covariance matrix.
    # The with upper triangular elements are vectorized in radiant: 'qrr', 'qrp', 'qry', 'qpp', 'qpy', 'qyy'
    Sigma_R_vec = None

    # correlation between position (p) and orientation (R): 3x3 covariance matrix.
    # The with upper triangular elements are vectorized: 'pxr', 'pxp', 'pxy', 'pyr', 'pyp', 'pyy', 'pzr', 'pzp', 'pzy'
    Sigma_pR_vec = None

    Sigma_T_vec = None

    # EstimationErrorType: specifies global or local definitions of the uncertainty (Sigma_p/q_vec)
    # ErrorRepresentationType: specifies the uncertainty of the rotation (Sigma_R_vec)
    format = CSVSpatialFormat()

    def __init__(self, t_vec=None, p_vec=None, q_vec=None,
                 Sigma_p_vec=None, Sigma_R_vec=None,
                 Sigma_pR_vec=None, Sigma_T_vec=None,
                 df=None, fn=None):
        Trajectory.__init__(self, t_vec=t_vec, p_vec=p_vec, q_vec=q_vec)
        self.Sigma_p_vec = Sigma_p_vec
        self.Sigma_R_vec = Sigma_R_vec
        self.Sigma_pR_vec = Sigma_pR_vec
        self.Sigma_T_vec = Sigma_T_vec

        if df is not None:
            self.load_from_DataFrame(df)
        elif fn is not None:
            self.load_from_CSV(fn)
        pass

    def set_format(self, fmt):
        if fmt is not None and isinstance(fmt, CSVSpatialFormat):
            self.format = fmt
        elif fmt is not None and isinstance(fmt, CSVSpatialFormatType):
            self.format = CSVSpatialFormat(fmt_type=fmt)

    def get_format(self):
        return self.format

    def convert_to_global_covariance(self):
        # Covariance of pose is expressed in the local frame
        if self.format.estimation_error_type == EstimationErrorType.type1:
            if self.Sigma_T_vec is not None:
                for i in range(self.num_elems()):
                    T_AB = SpatialConverter.p_q_HTMQ_to_SE3(self.p_vec[i,:], self.q_vec[i,:])
                    Adj_AB = base.tr2adjoint(T_AB)

                    # Sigma expressed in Navigation frame
                    Sigma_NB = self.Sigma_T_vec[i, :, :]
                    Sigma_GB = np.dot(Adj_AB, np.dot(Sigma_NB, Adj_AB.T))
                    self.Sigma_T_vec[i, :, :] = Sigma_GB

            self.format.estimation_error_type = EstimationErrorType.type2
        # Covariance of orientation is defined locally
        elif self.format.estimation_error_type == EstimationErrorType.type5:
            if self.Sigma_R_vec is not None:
                for i in range(self.num_elems()):
                    # Sigma expressed in Navigation frame
                    Sigma_q_NB = self.Sigma_R_vec[i, :, :]
                    R_GN = SpatialConverter.HTMQ_quaternion_to_rot(self.q_vec[i,:])
                    Sigma_q_GB = np.dot(R_GN, np.dot(Sigma_q_NB, R_GN.T))
                    self.Sigma_R_vec[i, :, :] = Sigma_q_NB
        # Covariance of position and orientation is defined globally
            self.format.estimation_error_type = EstimationErrorType.type6

        pass

    # overriding abstract method
    def transform(self, scale=1.0, p_GN_in_G=np.zeros((3,)), R_GN=np.identity(3)):
        # Calling the parent's class
        Trajectory.transform(self=self, scale=scale, p_GN_in_G=p_GN_in_G, R_GN=R_GN)

        # Covariance of pose is expressed in the global frame
        if self.format.estimation_error_type == EstimationErrorType.type2:
            if self.Sigma_T_vec is not None:
                T_AB = SpatialConverter.p_R_to_SE3(p_GN_in_G, R_GN)
                Adj_AB = base.tr2adjoint(T_AB)
                for i in range(self.num_elems()):
                    # Sigma expressed in Navigation frame
                    Sigma_NB = self.Sigma_T_vec[i, :, :]
                    Sigma_GB = np.dot(Adj_AB, np.dot(Sigma_NB, Adj_AB.T))
                    self.Sigma_T_vec[i, :, :] = Sigma_GB
        # Covariance of position is defined globally
        elif self.format.estimation_error_type == EstimationErrorType.type5:
            if self.Sigma_p_vec is not None:
                for i in range(self.num_elems()):
                    # Sigma expressed in Navigation frame
                    Sigma_p_NB = self.Sigma_p_vec[i, :, :]
                    Sigma_p_GB = np.dot(R_GN, np.dot(Sigma_p_NB, R_GN.T))
                    self.Sigma_p_vec[i, :, :] = Sigma_p_GB
        # Covariance of position and orientation is defined globally
        elif self.format.estimation_error_type == EstimationErrorType.type6:
            if self.Sigma_p_vec is not None and self.Sigma_R_vec is not None:
                for i in range(self.num_elems()):
                    # Sigma expressed in Navigation frame
                    Sigma_p_NB = self.Sigma_p_vec[i, :, :]
                    Sigma_p_GB = np.dot(R_GN, np.dot(Sigma_p_NB, R_GN.T))
                    self.Sigma_p_vec[i, :, :] = Sigma_p_GB

                    Sigma_R_NB = self.Sigma_R_vec[i, :, :]
                    Sigma_R_GB = np.dot(R_GN, np.dot(Sigma_R_NB, R_GN.T))
                    self.Sigma_p_vec[i, :, :] = Sigma_R_GB
        # else:
        #    local covariances are invariant to global reference changes!

    ########################################################

    # overriding abstract method
    def subsample(self, step=None, num_max_points=None, verbose=False):
        sparse_indices = Trajectory.subsample(self, step=step, num_max_points=num_max_points, verbose=verbose)

        if self.Sigma_p_vec is not None:
            self.Sigma_p_vec = self.Sigma_p_vec[sparse_indices, :, :]
        if self.Sigma_R_vec is not None:
            self.Sigma_R_vec = self.Sigma_R_vec[sparse_indices, :, :]
        if self.Sigma_pR_vec is not None:
            self.Sigma_pR_vec = self.Sigma_pR_vec[sparse_indices, :, :]
        if self.Sigma_T_vec is not None:
            self.Sigma_T_vec = self.Sigma_T_vec[sparse_indices, :, :]

        return sparse_indices

    # overriding abstract method
    def sample(self, indices_arr, verbose=False):
        Trajectory.sample(self, indices_arr=indices_arr)
        if self.Sigma_p_vec is not None:
            self.Sigma_p_vec = self.Sigma_p_vec[indices_arr, :, :]
        if self.Sigma_R_vec is not None:
            self.Sigma_R_vec = self.Sigma_R_vec[indices_arr, :, :]
        if self.Sigma_pR_vec is not None:
            self.Sigma_pR_vec = self.Sigma_pR_vec[indices_arr, :, :]
        if self.Sigma_T_vec is not None:
            self.Sigma_T_vec = self.Sigma_T_vec[indices_arr, :, :]

    # overriding abstract method
    def clone(self):
        obj =  TrajectoryEstimated(t_vec=self.t_vec.copy(), p_vec=self.p_vec.copy(), q_vec=self.q_vec.copy())
        obj.set_format(self.get_format())
        if self.Sigma_p_vec is not None:
            obj.Sigma_p_vec = self.Sigma_p_vec.copy()
        if self.Sigma_R_vec is not None:
            obj.Sigma_R_vec = self.Sigma_R_vec.copy()
        if self.Sigma_pR_vec is not None:
            obj.Sigma_pR_vec = self.Sigma_pR_vec.copy()
        if self.Sigma_T_vec is not None:
            obj.Sigma_T_vec = self.Sigma_T_vec.copy()
        return obj

    # overriding abstract method
    def load_from_DataFrame(self, df, fmt_type=None):
        assert (isinstance(df, pandas.DataFrame))
        if fmt_type is None:
            fmt_type = CSV2DataFrame.identify_format(dataframe=df)
        else:
            assert (isinstance(fmt_type, CSVSpatialFormatType))

        err_rep_type = ErrorRepresentationType.none
        est_err_type = EstimationErrorType.none
        # TODO: different uncertainties can be loaded! either full pose covariance or position and orientation separated
        if fmt_type == CSVSpatialFormatType.PosOrientWithCov:
            self.t_vec, self.p_vec, self.q_vec, self.Sigma_p_vec, self.Sigma_R_vec = \
                PosOrientWithCov2DataFrame.from_DataFrame(data_frame=df)
        elif fmt_type == CSVSpatialFormatType.PoseWithCov:
            self.t_vec, self.p_vec, self.q_vec, self.Sigma_T_vec = \
                PoseWithCov2DataFrame.from_DataFrame(data_frame=df)
        elif fmt_type == CSVSpatialFormatType.PoseStamped or fmt_type == CSVSpatialFormatType.TUM:
            self.t_vec, self.p_vec, self.q_vec = TUMCSV2DataFrame.from_DataFrame(data_frame=df)
        elif fmt_type == CSVSpatialFormatType.PosOrientWithCovTyped:
            self.t_vec, self.p_vec, self.q_vec, self.Sigma_p_vec, self.Sigma_R_vec, \
            est_err_type_vec, err_rep_vec = PosOrientWithCovTyped2DataFrame.from_DataFrame(data_frame=df)
            est_err_type = EstimationErrorType(est_err_type_vec[0])
            err_rep_type = ErrorRepresentationType(err_rep_vec[0])
        elif fmt_type == CSVSpatialFormatType.PoseWithCovTyped:
            self.t_vec, self.p_vec, self.q_vec, self.Sigma_T_vec, \
            est_err_type_vec, err_rep_vec = PoseWithCovTyped2DataFrame.from_DataFrame(data_frame=df)
            est_err_type = EstimationErrorType(est_err_type_vec[0])
            err_rep_type = ErrorRepresentationType(err_rep_vec[0])
        else:
            print('Error: format not supported yet')
            return False

        self.set_format(CSVSpatialFormat(fmt_type, est_err_type=est_err_type, err_rep_type=err_rep_type))
        return True

    # overriding abstract method
    def to_DataFrame(self):
        if self.format.type == CSVSpatialFormatType.PosOrientWithCov:
            return PosOrientWithCov2DataFrame.to_DataFrame(self.t_vec, self.p_vec, self.q_vec, self.Sigma_p_vec,
                                                           self.Sigma_R_vec)
        elif self.format.type == CSVSpatialFormatType.PoseWithCov:
            return PoseWithCov2DataFrame.to_DataFrame(self.t_vec, self.p_vec, self.q_vec, self.Sigma_T_vec)
        elif self.format.type == CSVSpatialFormatType.TUM:
            return TUMCSV2DataFrame.to_DataFrame(self.t_vec, self.p_vec, self.q_vec)
        elif self.format.type == CSVSpatialFormatType.PosOrientWithCovTyped:
            est_err_type_vec = np.repeat(self.format.estimation_error_type.str(), self.num_elems(), axis=0)
            err_rep_vec = np.repeat(self.format.rotation_error_representation.str(), self.num_elems(), axis=0)

            return PosOrientWithCovTyped2DataFrame.to_DataFrame(self.t_vec, self.p_vec, self.q_vec, self.Sigma_p_vec,
                                                                self.Sigma_R_vec,
                                                                est_err_type_vec=est_err_type_vec,
                                                                err_rep_vec=err_rep_vec)
        elif self.format.type == CSVSpatialFormatType.PoseWithCovTyped:
            est_err_type_vec = np.repeat(self.format.estimation_error_type.str(), self.num_elems(), axis=0)
            err_rep_vec = np.repeat(self.format.rotation_error_representation.str(), self.num_elems(), axis=0)

            return PoseWithCovTyped2DataFrame.to_DataFrame(self.t_vec, self.p_vec, self.q_vec, self.Sigma_T_vec,
                                                           est_err_type_vec=est_err_type_vec,
                                                           err_rep_vec=err_rep_vec)
        else:
            print('Error: format [' + str(self.format.type) + '] not supported yet')
            return pandas.DataFrame()

    ####################################################################################################################
    ######### PLOTTING #################################################################################################
    ####################################################################################################################
    def get_p_N_sigma_data(self, cfg, sigma_N=3.0):
        assert (isinstance(cfg, TrajectoryPlotConfig))

        negative_variance_detected = False

        sigma_p_diag_vec = np.zeros(np.shape(self.p_vec))
        if self.format.type == CSVSpatialFormatType.PosOrientWithCov or self.format.type == CSVSpatialFormatType.PosOrientWithCovTyped:
            for i in range(self.num_elems()):
                P = self.Sigma_p_vec[i]
                d_variance = np.diag(P)
                if min(d_variance) < 0:
                    negative_variance_detected = True
                    d_variance = abs(d_variance)

                sigma_p_diag_vec[i] = np.sqrt(d_variance)

        elif self.format.type == CSVSpatialFormatType.PoseWithCov or self.format.type == CSVSpatialFormatType.PoseWithCovTyped :
            for i in range(self.num_elems()):
                P = self.Sigma_T_vec[i]

                d_variance = np.diag(P[[0,1,2], [0,1,2]])
                if min(d_variance) < 0:
                    negative_variance_detected = True
                    d_variance = abs(d_variance)

                sigma_p_diag_vec[i] = np.sqrt(d_variance)
        else:
            print('format ' + str(self.format.type) + ' not supported!')
            assert False
        sigma_N_diag_vec = sigma_p_diag_vec * sigma_N

        if negative_variance_detected:
            print('WARNING: TrajectoryEstimated.get_p_N_sigma_data(): negative variance detected!')

        # sigma3_diag_vec = np.zeros(np.shape(self.p_vec))
        # if self.format.rotation_error_representation == ErrorRepresentationType.theta_R:
        #     for i in range(self.num_elems()):
        #         sigma3_diag_vec = sigma_p_diag_vec[i]
        # if not cfg.radians:
        #     sigma3_diag_vec = np.rad2deg(sigma3_diag_vec)

        ts = self.t_vec
        xs = sigma_N_diag_vec[:, 0]
        ys = sigma_N_diag_vec[:, 1]
        zs = sigma_N_diag_vec[:, 2]
        dist_vec = self.get_accumulated_distances()


        if cfg.white_list:
            print("white list args: " + str(cfg.white_list))
        if any([flag == 'x' for flag in cfg.white_list]):
            xs = []
            print("clear xs")
        if any([flag == 'y' for flag in cfg.white_list]):
            ys = []
            print("clear ys")
        if any([flag == 'z' for flag in cfg.white_list]):
            zs = []
            print("clear zs")

        if not (len(xs) and isinstance(xs[0], np.float64) and not math.isnan(xs[0])):
            xs = []
        if not (len(ys) and isinstance(ys[0], np.float64) and not math.isnan(ys[0])):
            ys = []
        if not (len(zs) and isinstance(zs[0], np.float64) and not math.isnan(zs[0])):
            zs = []

        return ts, xs, ys, zs, dist_vec

    def get_rpy_N_sigma_data(self, cfg, sigma_N=3.0):
        assert (isinstance(cfg, TrajectoryPlotConfig))

        negative_variance_detected = False

        sigma_theta_diag_vec = np.zeros(np.shape(self.p_vec))
        if self.format.type == CSVSpatialFormatType.PosOrientWithCov or self.format.type == CSVSpatialFormatType.PosOrientWithCovTyped:
            for i in range(self.num_elems()):
                d_variance = np.diag(self.Sigma_R_vec[i])
                if min(d_variance) < 0:
                    negative_variance_detected = True
                    d_variance = abs(d_variance)
                sigma_theta_diag_vec[i] = np.sqrt(d_variance)
        elif self.format.type == CSVSpatialFormatType.PoseWithCov or self.format.type == CSVSpatialFormatType.PoseWithCovTyped:
            for i in range(self.num_elems()):
                P = self.Sigma_T_vec[i]
                d_variance = np.diag(P[[3,4,5], [3,4,5]])
                if min(d_variance) < 0:
                    negative_variance_detected = True
                    d_variance = abs(d_variance)
                sigma_theta_diag_vec[i] = np.sqrt(d_variance)

        else:
            print('format ' + str(self.format.type) + ' not supported!')
            assert False

        if negative_variance_detected:
            print('WARNING: TrajectoryEstimated.get_rpy_N_sigma_data(): negative variance detected!')

        sigma_rpy_diag_vec = np.zeros(np.shape(self.p_vec))

        unit='rad'
        if not cfg.radians:
            unit='deg'

        # Converts small angle approximations of covariance to rpy angles: R =  Rz(y)Ry(p)Rx(r)
        if self.format.rotation_error_representation == ErrorRepresentationType.theta_R:
            for i in range(self.num_elems()):
                R = SpatialConverter.theta_R2rot(sigma_theta_diag_vec[i])
                sigma_rpy_diag_vec[i] = SpatialConverter.rot2rpy(R, unit=unit)
        elif self.format.rotation_error_representation == ErrorRepresentationType.theta_q:
            for i in range(self.num_elems()):
                R = SpatialConverter.theta_q2rot(sigma_theta_diag_vec[i])
                sigma_rpy_diag_vec[i] = SpatialConverter.rot2rpy(R, unit=unit)
        elif self.format.rotation_error_representation == ErrorRepresentationType.theta_so3:
            for i in range(self.num_elems()):
                R = SpatialConverter.theta_so3_2rot(sigma_theta_diag_vec[i])
                sigma_rpy_diag_vec[i] = SpatialConverter.rot2rpy(R, unit=unit)
        else:
            print('format is not supported: ' + str(self.format.rotation_error_representation))
            assert False

        sigma_N_diag_vec = sigma_rpy_diag_vec * sigma_N

        ts = self.t_vec
        xs = sigma_N_diag_vec[:, 0]
        ys = sigma_N_diag_vec[:, 1]
        zs = sigma_N_diag_vec[:, 2]
        dist_vec = self.get_accumulated_distances()

        if cfg.white_list:
            print("white list args: " + str(cfg.white_list))
        if any([flag == 'x' for flag in cfg.white_list]):
            xs = []
            print("clear xs")
        if any([flag == 'y' for flag in cfg.white_list]):
            ys = []
            print("clear ys")
        if any([flag == 'z' for flag in cfg.white_list]):
            zs = []
            print("clear zs")

        if not (len(xs) and isinstance(xs[0], np.float64) and not math.isnan(xs[0])):
            xs = []
        if not (len(ys) and isinstance(ys[0], np.float64) and not math.isnan(ys[0])):
            ys = []
        if not (len(zs) and isinstance(zs[0], np.float64) and not math.isnan(zs[0])):
            zs = []

        return ts, xs, ys, zs, dist_vec

    def ax_plot_rpy_sigma(self, ax, cfg,
                          sigma_N=3.0,
                          x_label_prefix='',
                          y_label_prefix='sigma rpy-zyx',
                          colors=['r', 'g', 'b'],
                          labels=['x', 'y', 'z'],
                          ls=PlotLineStyle()):
        assert (isinstance(cfg, TrajectoryPlotConfig))

        ts, xs, ys, zs, dist_vec = self.get_rpy_N_sigma_data(cfg, sigma_N=sigma_N)

        TrajectoryPlotUtils.ax_plot_rpy(ax=ax,ts=ts, xs=xs, ys=ys, zs=zs, dist_vec=dist_vec,
                                        x_label_prefix=x_label_prefix, y_label_prefix=str(sigma_N)+y_label_prefix,
                                        plot_type=cfg.plot_type,
                                        radians=cfg.radians,
                                        relative_time=cfg.relative_time,
                                        colors=colors, labels=labels, ls=ls)
        TrajectoryPlotUtils.ax_plot_rpy(ax=ax,ts=ts, xs=-xs, ys=-ys, zs=-zs, dist_vec=dist_vec,
                                        x_label_prefix=x_label_prefix, y_label_prefix=str(sigma_N)+y_label_prefix,
                                        plot_type=cfg.plot_type,
                                        radians=cfg.radians,
                                        relative_time=cfg.relative_time,
                                        colors=colors, labels=labels, ls=ls)

    def ax_plot_p_sigma(self, ax, cfg,
                          sigma_N=3.0,
                          x_label_prefix='',
                          y_label_prefix='sigma ',
                          colors=['r', 'g', 'b'],
                          labels=['x', 'y', 'z'],
                          ls=PlotLineStyle()):
        assert (isinstance(cfg, TrajectoryPlotConfig))

        ts, xs, ys, zs, dist_vec = self.get_p_N_sigma_data(cfg, sigma_N=sigma_N)

        TrajectoryPlotUtils.ax_plot_pos(ax=ax, ts=ts, xs=xs, ys=ys, zs=zs, dist_vec=dist_vec,
                                        x_label_prefix=x_label_prefix, y_label_prefix=str(sigma_N)+y_label_prefix,
                                        plot_type=cfg.plot_type,
                                        relative_time=cfg.relative_time,
                                        colors=colors, labels=labels, ls=ls)
        TrajectoryPlotUtils.ax_plot_pos(ax=ax, ts=ts, xs=-xs, ys=-ys, zs=-zs, dist_vec=dist_vec,
                                        x_label_prefix=x_label_prefix, y_label_prefix=str(sigma_N)+y_label_prefix,
                                        plot_type=cfg.plot_type,
                                        relative_time=cfg.relative_time,
                                        colors=colors, labels=labels, ls=ls)

    def plot_pos_orient_cov(self, cfg=TrajectoryPlotConfig(), fig=None, angles=False, plot_angle=False):
        assert (isinstance(cfg, TrajectoryPlotConfig))

        if fig is None:
            fig = plt.figure(figsize=(20, 15), dpi=int(cfg.dpi))

        # create 2x2 grid
        ax1 = fig.add_subplot(221)
        ax2 = fig.add_subplot(222)
        ax3 = fig.add_subplot(223)
        ax4 = fig.add_subplot(224)

        self.ax_plot_pos(ax=ax1, cfg=cfg)

        if plot_angle:
            self.ax_plot_angle(ax=ax2, cfg=cfg)
        else:
            if angles:
                self.ax_plot_rpy(ax=ax2, cfg=cfg)
            else:
                self.ax_plot_q(ax=ax2, cfg=cfg)

        self.ax_plot_p_sigma(ax3, cfg=cfg, colors=['darkred', 'darkgreen', 'darkblue'])
        self.ax_plot_rpy_sigma(ax4, cfg=cfg, colors=['darkred', 'darkgreen', 'darkblue'])

        TrajectoryPlotConfig.show_save_figure(cfg, fig)
        return fig, ax1, ax2, ax3, ax4


#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2020, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################################################################

import time
import os
import argparse
import yaml
import csv
from tqdm import tqdm

from cnspy_trajectory import TrajectoryPlotTypes
from cnspy_trajectory.Trajectory import Trajectory
from cnspy_trajectory.TrajectoryPlotConfig import TrajectoryPlotConfig
from cnspy_trajectory.TrajectoryPlotTypes import TrajectoryPlotTypes
from cnspy_trajectory.TrajectoryPlotter import TrajectoryPlotter


class PlotTrajectory:
    def __init__(self):
        pass

    @staticmethod
    def plot_single(fn, result_dir=None, verbose=False, plot_3D=True, plot_pose=True, show_plots=True, save_metrics=True):
        if not os.path.isfile(fn):
            print("PlotTrajectory: could not find file: %s" % fn)
            return False

        ## create result dir:
        if result_dir is None:
            folder = str(fn).replace(".csv", "")
        else:
            folder = result_dir
        folder = os.path.abspath(folder)
        try:  # else already exists
            os.makedirs(folder)
        except:
            pass
        if verbose:
            print("* result_dir: \t " + str(folder))

        traj = Trajectory()
        traj.load_from_CSV(fn=fn)
        if not traj.is_empty():
            if plot_3D:
                cfg = TrajectoryPlotConfig(show=show_plots, close_figure=True,
                                           save_fn=str(folder + '/plot_3D.png'))
                traj.plot_3D(cfg=cfg)
            if plot_pose:
                cfg=TrajectoryPlotConfig(show=show_plots,
                                         close_figure=True,
                                         radians=True,
                                         plot_type=TrajectoryPlotTypes.plot_2D_over_t,
                                         save_fn=str(folder + '/pose_over_t.png'))
                traj.plot_pose(cfg=cfg)
                cfg.plot_type=TrajectoryPlotTypes.plot_2D_over_dist
                cfg.save_fn=str(folder + '/pose_over_dist.png')
                traj.plot_pose(cfg=cfg)

            if save_metrics:
                if verbose:
                    traj.print_statistics()
                with open(folder + '/metrics.txt', 'w') as file:
                    traj.print_statistics(file=file)

        else:
            print("* Trajectory was not loaded!")
            return False
        return True

    @staticmethod
    def plot_multi(fns, result_dir=None, verbose=False, show_plots=True):
        folder = result_dir
        traj_arr = []
        for fn in fns:
            if not os.path.isfile(fn):
                print("* WARNING: could not find file: %s" % fn)
                continue
            ## create result dir:
            if folder is None:
                folder = str(fn).replace(".csv", "")
                folder = os.path.abspath(folder)
                try:  # else already exists
                    [head, tail] = os.path.split(folder)
                    folder = head
                    os.makedirs(folder)
                except:
                    pass
                if verbose:
                    print("* result_dir: \t " + str(folder))

            traj = Trajectory()
            traj.load_from_CSV(fn=fn)
            if not traj.is_empty():
                traj_arr.append(traj)
            else:
                print("* WARNING trajectory is empty! fn=" + str(fn))

        if traj_arr:
            cfg = TrajectoryPlotConfig(show=show_plots, close_figure=True,
                                       save_fn=str(folder + '/multiplot_3D.png'))
            TrajectoryPlotter.multi_plot_3D(traj_arr, cfg=cfg)

        return True


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='PlotTrajectory: tool to save plots and metrics of a trajectory')
    parser.add_argument('--filename', help='csv filename', default=None)
    parser.add_argument('--filenames', type=str, nargs='+',
                        help='a list of files to be plotted in 3D', default=None)
    parser.add_argument('--result_dir', help='directory to store results]',
                        default=None)
    parser.add_argument('--verbose', action='store_true', default=False)
    parser.add_argument('--plot_3D', action='store_true', default=False)
    parser.add_argument('--plot_pose', action='store_true', default=False)
    parser.add_argument('--show_plots', action='store_true', default=False)
    parser.add_argument('--save_metrics', action='store_true', default=False)


    tp_start = time.time()
    args = parser.parse_args()

    if args.filename is not None:
        PlotTrajectory.plot_single(args.filename, result_dir=args.result_dir, verbose=args.verbose,
                               plot_3D=args.plot_3D, plot_pose=args.plot_pose, show_plots=args.show_plots,
                               save_metrics=args.save_metrics)

    elif args.filenames is not None:
        PlotTrajectory.plot_multi(args.filenames, result_dir=args.result_dir, show_plots=args.show_plots, verbose=args.verbose)

    print("finished after [%s sec]\n" % str(time.time() - tp_start))



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2020, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Requirements:
# numpy, matplotlib, cnspy_csv2dataframe, cnspy_trajectory
########################################################################################################################
import os
import unittest
import math

from cnspy_spatial_csv_formats.CSVSpatialFormat import CSVSpatialFormat
from cnspy_spatial_csv_formats.CSVSpatialFormatType import CSVSpatialFormatType
from cnspy_spatial_csv_formats.ErrorRepresentationType import ErrorRepresentationType
from cnspy_spatial_csv_formats.EstimationErrorType import EstimationErrorType
from cnspy_trajectory.TrajectoryErrorType import TrajectoryErrorType
from cnspy_trajectory.TrajectoryPlotter import *

SAMPLE_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'sample_data')

class TrajectoryPlotter_Test(unittest.TestCase):
    def load_trajectory_gt_from_CSV(self):
        traj = Trajectory()
        traj.load_from_CSV(fn=str(SAMPLE_DATA_DIR + '/ID1-pose-gt.csv'))
        self.assertFalse(traj.is_empty())
        return traj

    def load_trajectory_posorient_from_CSV(self):
        traj = TrajectoryEstimated()
        traj.load_from_CSV(fn=str(SAMPLE_DATA_DIR + '/ID1-pose-est-posorient-cov.csv'))
        self.assertFalse(traj.is_empty())
        return traj

    def load_gt_est_err(self):
        traj_gt = Trajectory()
        traj_gt.load_from_CSV(fn=str(SAMPLE_DATA_DIR + '/ID1-pose-gt.csv'))
        traj_est = TrajectoryEstimated()
        traj_est.load_from_CSV(fn=str(SAMPLE_DATA_DIR + '/ID1-pose-est-posorient-cov.csv'))
        traj_est.format = CSVSpatialFormat(fmt_type=CSVSpatialFormatType.PosOrientWithCov,
                                           est_err_type=EstimationErrorType.type5,
                                           err_rep_type=ErrorRepresentationType.theta_R)
        traj_err = TrajectoryError()
        traj_err.load_from_CSV(fn=str(SAMPLE_DATA_DIR + '/ID1-pose-err.csv'))
        traj_err.traj_err_type = TrajectoryErrorType(err_type=EstimationErrorType.type5)

        return traj_gt, traj_est, traj_err

    def test_plot_multi(self):

        traj_gt = self.load_trajectory_gt_from_CSV()
        traj_est = self.load_trajectory_posorient_from_CSV()
        cfg = TrajectoryPlotConfig(show=True,
                                   save_fn=str(SAMPLE_DATA_DIR + '/../../doc/multi.png'),
                                   view_angle=(40, 20))
        TrajectoryPlotter.multi_plot_3D([traj_gt, traj_est], cfg=cfg, name_list=['gt', 'est'])

    def test_full_error_plot(self):
        traj_gt, traj_est, traj_err = self.load_gt_est_err()
        cfg = TrajectoryPlotConfig(show=True,
                                   radians=False,
                                   close_figure=False,
                                   save_fn=str(SAMPLE_DATA_DIR + '/../../doc/traj_plotter_full_err.png'),
                                   view_angle=(40, 20))

        TrajectoryPlotter.plot_pose_err_cov(traj_est=traj_est, traj_err=traj_err, traj_gt=traj_gt, cfg=cfg)




if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2020, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Requirements:
# enum
########################################################################################################################
import os

from cnspy_spatial_csv_formats.CSVSpatialFormatType import CSVSpatialFormatType
from cnspy_spatial_csv_formats.ErrorRepresentationType import ErrorRepresentationType
from cnspy_trajectory.Trajectory import Trajectory
from cnspy_trajectory.TrajectoryEstimated import TrajectoryEstimated
import unittest
import time

from cnspy_trajectory.TrajectoryPlotConfig import TrajectoryPlotConfig
from cnspy_trajectory.TrajectoryPlotTypes import TrajectoryPlotTypes

SAMPLE_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'sample_data')

class TrajectoryEstimated_Test(unittest.TestCase):
    start_time = None

    def start(self):
        self.start_time = time.time()

    def stop(self):
        print("Process time: " + str((time.time() - self.start_time)))

    def load_(self):
        print('loading...')
        obj = TrajectoryEstimated(fn = str(SAMPLE_DATA_DIR + '/ID1-pose-est-posorient-cov.csv'))
        obj.format.rotation_error_representation = ErrorRepresentationType.theta_R
        return obj

    def test_load_trajectory_from_CSV(self):
        self.start()
        obj = self.load_()
        # self.assertTrue(obj.data_loaded)
        self.stop()

        print('\n' + str(obj.Sigma_p_vec[1000]))
        print('\n' + str(obj.Sigma_p_vec[1000]))
        self.start()
        obj.save_to_CSV(fn=str(SAMPLE_DATA_DIR + '/results/ID1-pose-est-cov-copy.csv'))
        obj.format.type = CSVSpatialFormatType.PosOrientWithCovTyped
        obj.save_to_CSV(fn=str(SAMPLE_DATA_DIR + '/results/ID1-pose-est-cov-typed-copy.csv'))
        self.stop()

    def test_load_pose_cov_traj(self):
        fn = str(SAMPLE_DATA_DIR + '/ID1-pose-est-pose-cov.csv')
        obj = TrajectoryEstimated()
        obj.load_from_CSV(fn=fn)
        print('\n' + str(obj.Sigma_T_vec[1]))

    def test_plot_pos_orient_cov_theta_R(self):
        traj_est = self.load_()
        traj_est.format.rotation_error_representation = ErrorRepresentationType.theta_R
        cfg = TrajectoryPlotConfig(show=False, close_figure=False, radians=False,
                                   plot_type=TrajectoryPlotTypes.plot_2D_over_dist,
                                   save_fn=str(SAMPLE_DATA_DIR + '/../../doc/traj_est_pos_orient_cov_theta_R.png'))
        traj_est.plot_pos_orient_cov(angles=True, cfg=cfg)


    def test_plot_pos_orient_cov_theta_q(self):
        traj_est = self.load_()
        traj_est.format.rotation_error_representation = ErrorRepresentationType.theta_q
        cfg = TrajectoryPlotConfig(show=False, close_figure=False,
                                   save_fn=str(SAMPLE_DATA_DIR + '/../../doc/traj_est_pos_orient_cov_theta_q.png'))
        traj_est.plot_pos_orient_cov(angles=True, cfg=cfg)

    def test_plot_pos_orient_cov_theta_so3(self):
        traj_est = self.load_()
        traj_est.format.rotation_error_representation = ErrorRepresentationType.theta_so3
        cfg = TrajectoryPlotConfig(show=False, close_figure=False, radians=False,
                                   save_fn=str(SAMPLE_DATA_DIR + '/../../doc/traj_est_pos_orient_cov_theta_so3.png'))
        traj_est.plot_pos_orient_cov(angles=True, cfg=cfg)

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2022, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################################################################
import math
import os
from sys import version_info
from abc import ABC, abstractmethod
import numpy as np
import pandas

from cnspy_csv2dataframe.CSV2DataFrame import CSV2DataFrame
from cnspy_timestamp_association.TimestampAssociation import TimestampAssociation


# abstract methods:
# - to_DataFrame(self):
# - load_from_DataFrame(self, df, fmt_type=None)
class TrajectoryBase(ABC):
    t_vec = None

    @abstractmethod
    def to_DataFrame(self):
        pass

    @abstractmethod
    def load_from_DataFrame(self, df, fmt_type=None):
        pass

    @abstractmethod
    def subsample(self, step=None, num_max_points=None, verbose=False):
        num_elems = self.num_elems()

        if num_max_points:
            step = 1
            if (int(num_max_points) > 0) and (int(num_max_points) < num_elems):
                step = int(math.ceil(num_elems / float(num_max_points)))

        sparse_indices = np.arange(start=0, stop=num_elems, step=step)
        if (num_max_points or step):
            if verbose:
                print("TrajectoryBase.subsample():")
                print("* len: " + str(num_elems) + ", max_num_points: " + str(
                    num_max_points) + ", subsample by: " + str(step))

            self.t_vec = self.t_vec[sparse_indices]

        return sparse_indices

    @abstractmethod
    def sample(self, indices_arr, verbose=False):
        num_elems = self.num_elems()
        assert (len(indices_arr) <= num_elems), "TrajectoryBase.sample():\n\t index array must be smaller " \
                                                "equal the dataframe."
        assert (max(indices_arr) <= num_elems), "TrajectoryBase.sample():\n\t elemts in the index array " \
                                                "must be smaller equal the dataframe."

        assert (min(indices_arr) >= 0), "TrajectoryBase.sample():\n\t elemts in the index array " \
                                        "must be greater equal zero."
        if verbose:
            print("TrajectoryBase.sample():")
            print("* current len: " + str(num_elems) + ", take N-samples: " + len(indices_arr))

        self.t_vec = self.t_vec[indices_arr]

    @abstractmethod
    def clone(self):
        pass


    def is_empty(self):
        return self.num_elems() == 0

    def num_elems(self):
        if self.t_vec is None:
            return 0
        else:
            return len(self.t_vec)

    def sample_at_t_arr(self, t_arr, max_difference=0, round_decimals=9, unique_timestamps=False):
        indices_arr, idx_t, t_vec_matched, t_arr_matched = \
            TimestampAssociation.associate_timestamps(self.t_vec, t_arr,
                                                      max_difference=max_difference,
                                                      round_decimals=round_decimals,
                                                      unique_timestamps=unique_timestamps)
        # call abstract method
        self.sample(indices_arr)
        return idx_t, t_arr_matched

    def load_from_CSV(self, fn):
        if not os.path.isfile(fn):
            print("TrajectoryBase: could not find file %s" % os.path.abspath(fn))
            return False
        try:
            loader = CSV2DataFrame(fn=fn)
            self.load_from_DataFrame(loader.data_frame, fmt_type=loader.format)
            return loader.data_loaded
        except:
            print("TrajectoryBase: could not load file %s" % os.path.abspath(fn))
            return False

    def save_to_CSV(self, fn):
        if self.is_empty():
            return False
        CSV2DataFrame.save_CSV(self.to_DataFrame(), filename=fn)
        return True



#!/usr/bin/env python

from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='cnspy_trajectory',
    version="0.2.1",
    author='Roland Jung',
    author_email='roland.jung@aau.at',    
    description='Trajectory in SE(3) space with utilities.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/aau-cns/cnspy_trajectory/',
    project_urls={
        "Bug Tracker": "https://github.com/aau-cns/cnspy_trajectory/issues",
    },    
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    
    packages=find_packages(exclude=["test_*", "TODO*"]),
    python_requires='>=3.6',
    install_requires=['numpy', 'matplotlib', 'cnspy_csv2dataframe', 'spatialmath-python', 'cnspy_spatial_csv_formats'],
)



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2022, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################################################################
import os
from sys import version_info

import numpy as np
import pandas

from cnspy_csv2dataframe.CSV2DataFrame import CSV2DataFrame
from cnspy_csv2dataframe.PoseErrorStamped2DataFrame import PoseErrorStamped2DataFrame
from cnspy_csv2dataframe.TUMCSV2DataFrame import TUMCSV2DataFrame
from cnspy_spatial_csv_formats.CSVSpatialFormatType import CSVSpatialFormatType
from cnspy_spatial_csv_formats.ErrorRepresentationType import ErrorRepresentationType
from cnspy_spatial_csv_formats.EstimationErrorType import EstimationErrorType
from cnspy_trajectory.TrajectoryBase import TrajectoryBase


# - TODO: save and load to CSV file and DataFrame: introduce new format!
class TrajectoryEstimationError(TrajectoryBase):
    # p_vec: position error vector defined via "error_type" [m]
    # q_vec: rotation error defined via "error_type"  [quaternion]
    nu_vec = None     # position error converted into space and unit of the rotation covariance matrix
    theta_vec = None  # rotation error converted into space and unit of the rotation covariance matrix
    est_err_type = EstimationErrorType.none # EstimationErrorType
    err_rep_type = ErrorRepresentationType.none # ErrorRepresentationType

    def __init__(self, t_vec=None, nu_vec=None, theta_vec=None,
                 est_err_type=EstimationErrorType.none, err_rep_type=ErrorRepresentationType.none,
                 df=None, fn=None):
        TrajectoryBase.__init__(self)
        if df is not None:
            self.load_from_DataFrame(df)
        elif fn is not None:
            self.load_from_CSV(fn)
        elif t_vec is not None and nu_vec is not None and theta_vec is not None:
            if t_vec.ndim == 1:
                t_vec = np.array([t_vec])

            t_rows, t_cols = t_vec.shape
            nu_rows, nu_cols = nu_vec.shape
            theta_rows, theta_cols = theta_vec.shape
            assert (t_rows == nu_rows)
            assert (t_rows == theta_rows)
            assert (theta_cols == 3)
            assert (nu_cols == 3)

            self.nu_vec = nu_vec
            self.theta_vec = theta_vec
            self.t_vec = t_vec
            self.est_err_type = est_err_type
            self.err_rep_type = err_rep_type
        pass

    # overriding abstract method
    def subsample(self, step=None, num_max_points=None, verbose=False):
        sparse_indices = TrajectoryBase.subsample(self, step=step, num_max_points=num_max_points, verbose=verbose)
        self.nu_vec = self.nu_vec[sparse_indices]
        self.theta_vec = self.theta_vec[sparse_indices]

        return sparse_indices

    # overriding abstract method
    def sample(self, indices_arr, verbose=False):
        TrajectoryBase.sample(self, indices_arr=indices_arr)
        self.nu_vec = self.nu_vec[indices_arr]
        self.theta_vec = self.theta_vec[indices_arr]

    # overriding abstract method
    def clone(self):
        obj = TrajectoryEstimationError(t_vec=self.t_vec.copy(),
                                        nu_vec=self.nu_vec.copy(),
                                        theta_vec=self.theta_vec.copy(),
                                        est_err_type=self.est_err_type,
                                        err_rep_type=self.err_rep_type)
        return obj

    # overriding abstract method
    def load_from_DataFrame(self, df, fmt_type=None):
        assert (isinstance(df, pandas.DataFrame))
        if fmt_type is None:
            fmt_type = CSV2DataFrame.identify_format(dataframe=df)
        else:
            assert (isinstance(fmt_type, CSVSpatialFormatType))

        est_err_type = EstimationErrorType.none
        err_rep_type = ErrorRepresentationType.none
        if fmt_type == CSVSpatialFormatType.PoseErrorStamped:
            self.t_vec, self.nu_vec, self.theta_vec, est_err_type_vec, err_rep_vec = \
                PoseErrorStamped2DataFrame.from_DataFrame(df)
            est_err_type = EstimationErrorType(est_err_type_vec[0])
            err_rep_type = ErrorRepresentationType(err_rep_vec[0])

        self.est_err_type = est_err_type
        self.err_rep_type = err_rep_type

    # overriding abstract method
    def to_DataFrame(self):
        est_err_type_vec = np.tile(self.est_err_type.str(), (self.num_elems(), 1))
        err_rep_vec = np.tile(self.err_rep_type.str(), (self.num_elems(), 1))
        return PoseErrorStamped2DataFrame.to_DataFrame(self.t_vec, self.nu_vec, self.theta_vec,
                                                       est_err_type_vec=est_err_type_vec, err_rep_vec=err_rep_vec)



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2020, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Requirements:
########################################################################################################################

class PlotLineStyle():
    linecolor = 'r'  # b,g,r,c,m,y,k,w
    linewidth = 1
    linestyle = '-'  # '-', '--', '-.', ':'
    marker = '.'  # '.', ',', 'o', 'v', '^', '<', '>', '1..4', 's', 'p'
    markersize = 12
    markerfacecolor = 'blue'
    label = 'x'

    def __init__(self, linecolor='r', linewidth=1, linestyle='-', marker='o', markersize=12,
                 markerfacecolor='blue', label='x'):
        self.linecolor = linecolor
        self.linewidth = linewidth
        self.linestyle = linestyle
        self.marker = marker
        self.markersize = markersize
        self.markerfacecolor = markerfacecolor
        self.label = label



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2022, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################################################################
import os
import unittest
from cnspy_spatial_csv_formats.EstimationErrorType import EstimationErrorType
from cnspy_trajectory.TrajectoryErrorType import TrajectoryErrorType
SAMPLE_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'sample_data')

class TrajectoryErrorType_Test(unittest.TestCase):
    def test_list(self):
        err_type = TrajectoryErrorType(EstimationErrorType.type1)

        print('\n* type: ' + str(err_type))

        print("\n" + str(err_type.is_local_rotation_error()))
        print("\n" + str(err_type.error_def()))

        print("\n" + str(TrajectoryErrorType.list()))



if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2020, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Requirements:
# enum
########################################################################################################################
import os
import matplotlib.pyplot as plt
from cnspy_trajectory.TrajectoryPlotTypes import TrajectoryPlotTypes


# https://matplotlib.org/2.1.2/api/_as_gen/matplotlib.pyplot.plot.html
class TrajectoryPlotConfig():
    white_list = []
    num_points = 0
    plot_type = TrajectoryPlotTypes.plot_3D
    dpi = 200
    title = ""
    scale = 1.0
    save_fn = ""
    result_dir = "."
    show = True
    close_figure = False
    radians = True
    relative_time = True
    view_angle = (45, 45)  # viewing angle of 3D plots tuple (azimuth, elevation)

    def __init__(self, white_list=[], num_points=0,
                 plot_type=TrajectoryPlotTypes.plot_3D, dpi=200, title="",
                 scale=1.0, save_fn="", result_dir=".", show=True, close_figure=False, radians=True, relative_time=True,
                 view_angle=(45, 45)):
        self.white_list = white_list
        self.num_points = num_points
        self.plot_type = plot_type
        self.dpi = dpi
        self.title = title
        self.scale = scale
        self.save_fn = save_fn
        self.result_dir = result_dir
        self.show = show
        self.close_figure = close_figure
        self.radians = radians
        self.relative_time = relative_time
        self.view_angle = view_angle

    @staticmethod
    def set_view_angle(cfg, ax):
        """
        set the view angle of a 3D plot

        Input:
        cfg -- TrajectoryPlotConfig
        ax  -- axes.Axes
        """
        assert (isinstance(cfg, TrajectoryPlotConfig))
        ax.view_init(azim=cfg.view_angle[0], elev=cfg.view_angle[1])

    @staticmethod
    def show_save_figure(cfg, fig):
        assert (isinstance(cfg, TrajectoryPlotConfig))
        assert(isinstance(fig, plt.Figure))
        plt.pause(0.01)
        plt.draw()
        plt.pause(0.01)
        if cfg.save_fn:
            if not os.path.exists(cfg.result_dir):
                os.makedirs(cfg.result_dir)

            filename = os.path.join(cfg.result_dir, cfg.save_fn)
            print("save to file: " + filename)
            plt.savefig(filename, dpi=int(cfg.dpi))
        if cfg.show:
            plt.show()
        if cfg.close_figure:
            plt.close(fig)
